"""XACML CombinerParameter type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: combinerparameter.py 7101 2010-06-25 16:13:06Z pjkersha $"

class CombinerParameter(object):
    '''XACML Combiner parameter class - currently not implemented
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    '''
    ELEMENT_LOCAL_NAME = "CombinerParameter"
    __slots__ = ()

    def __init__(self):
        '''@raise NotImplementedError: this class is a stub only
        '''
        raise NotImplementedError('Not currently implemented')


"""NDG XACML ElementTree based Target Element reader 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: conditionreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.condition import Condition
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.applyreader import ApplyReader


class ConditionReader(ETreeAbstractReader):
    '''ElementTree based XACML 2.0 Condition parser.  Note the difference to
    XACML 1.0: the Condition element is its own type and not an Apply type.
    It expects a single Expression derived type child element
    
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type
    '''
    TYPE = Condition
    
    def __call__(self, obj):
        """Parse condition object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML condition instance
        @rtype: ndg.xacml.core.condition.Condition
        @raise XMLParseError: error reading sub-elements
        """
        elem = super(ConditionReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        condition = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Parse expression sub-element
        nSubElem = len(elem)
        if nSubElem != 1:
            raise XMLParseError('XACML 2.0 policy schema expects only one '
                                'expression sub-element in the Condition '
                                'element; policy file has %d' % nSubElem)
        
        subElemlocalName = QName.getLocalPart(elem[0].tag)
        if subElemlocalName == xacmlType.APPLY_ELEMENT_LOCAL_NAME:
            condition.expression = ApplyReader.parse(elem[0])
        else:
            raise XMLParseError('Expecting %r Condition sub-element not '
                                'recognised' %
                                xacmlType.EXPRESSION_ELEMENT_LOCAL_NAME)
            
        
   
        return condition


"""NDG XACML module for Request type 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "23/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: request.py 7336 2010-08-18 09:07:40Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.utils import TypedList
from ndg.xacml.core.context import XacmlContextBase
from ndg.xacml.core.context.subject import Subject
from ndg.xacml.core.context.resource import Resource
from ndg.xacml.core.context.action import Action
from ndg.xacml.core.context.environment import Environment
from ndg.xacml.core.context.handlerinterface import CtxHandlerInterface


class Request(XacmlContextBase):
    """XACML Request class
    
    @cvar ELEMENT_LOCAL_NAME: XML local element name, derived classes should
    set
    @type ELEMENT_LOCAL_NAME: string
    
    @ivar __subjects: list of subjects corresponding to this request
    @type __subjects: ndg.xacml.utils.TypedList
    @ivar __resources: list of resources corresponding to this request
    @type __subjects: ndg.xacml.utils.TypedList
    @ivar __action: action for this request 
    @type __action: None / ndg.xacml.core.context.action.Action
    @ivar __environment: environment associated with this request 
    @type __environment: None / ndg.xacml.core.context.environment.Environment
    
    @ivar ctxHandler: reference to context handler to enable the PDP to
    query for additional attributes.  The Context Handler itself queries a
    Policy Information Point.  This handler setting may be omitted.  If so,
    the PDP will rely entirely on the input request context for making 
    access control decisions
    @type ctxHandler: ndg.xacml.core.context.handler.CtxHandlerInterface / 
    None
    """
    __slots__ = (
        '__subjects', 
        '__resources', 
        '__action', 
        '__environment',
        '__ctxHandler',
    )
    ELEMENT_LOCAL_NAME = 'Request'
    
    def __init__(self):
        super(Request, self).__init__()
        
        self.__subjects = TypedList(Subject)
        self.__resources = TypedList(Resource)
        self.__action = None
        self.__environment = None
        
        self.__ctxHandler = None
                    
    @property
    def subjects(self):
        """Get Request subjects
        @return: list of subjects
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__subjects
        
    @property
    def resources(self):
        """Get Request resources
        @return: list of resources
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__resources
                                    
    @property
    def action(self):
        """Get Request action
        @return: action type
        @rtype: None / ndg.xacml.core.context.action.Action
        """
        return self.__action
    
    @action.setter
    def action(self, value):
        """Set Request action
        @param value: action type
        @type value: ndg.xacml.core.context.action.Action
        """
        if not isinstance(value, Action):
            raise TypeError('Expecting %r type for request "action" '
                            'attribute; got %r' % (Action, type(value)))
            
        self.__action = value
                                    
    @property
    def environment(self):
        """Get Request environment
        @return: environment settings
        @rtype: None / ndg.xacml.core.context.environment.Environment
        """
        return self.__environment
    
    @environment.setter
    def environment(self, value):
        """Set Request environment
        @param value: environment settings
        @type value: ndg.xacml.core.context.environment.Environment
        """
        if not isinstance(value, Environment):
            raise TypeError('Expecting %r type for request "environment" '
                            'attribute; got %r' % (Environment, type(value)))
             
        self.__environment = value   

    @property
    def ctxHandler(self):
        """Get Context handler used by evaluate method to query the PIP for
        additional attribute values
        @return: context handler
        @rtype: None / ndg.xacml.core.context.handler.CtxHandlerInterface 
        derived type  
        """
        return self.__ctxHandler

    @ctxHandler.setter
    def ctxHandler(self, value):
        """Set Context handler used by evaluate method to query the PIP for
        additional attribute values
        @param value: context handler
        @type value: ndg.xacml.core.context.handler.CtxHandlerInterface 
        derived type  
        """
        if not isinstance(value, CtxHandlerInterface):
            raise TypeError('Expecting %r type for "ctxHandler" attribute; got '
                            '%r' % (CtxHandlerInterface, type(value)))
            
        self.__ctxHandler = value


"""NDG XACML ElementTree based Subject Element reader 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: subjectreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.subject import Subject
from ndg.xacml.parsers.etree.targetchildreader import TargetChildReader
    

class SubjectReader(TargetChildReader):
    '''ElementTree based XACML Rule parser
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: type
    '''
    TYPE = Subject



"""NDG Security Context handler interface definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id:$"
from abc import ABCMeta, abstractmethod

       
class CtxHandlerInterface(object):
    """Context Handler interface."""
    __metaclass__ = ABCMeta
    __slots__ = ()
    
    @abstractmethod
    def handlePEPRequest(self, pepRequest):
        """Handle request from Policy Enforcement Point
        
        @param pepRequest: request from PEP, derived class determines its type
        e.g. SAML AuthzDecisionQuery
        @type pepRequest: type
        @return: PEP response - derived class determines type
        @rtype: None
        """
        raise NotImplementedError() 
       
    def pipQuery(self, request, designator):
        """Query a Policy Information Point to retrieve the attribute values
        corresponding to the specified input designator.  Optionally, update the
        requestCtx.  This could be a subject, environment or resource.  Matching
        attributes values are returned
        
        @param request: request context
        @type request: ndg.xacml.core.context.request.Request
        @param designator: designator requiring additional subject attribute 
        information
        @type designator: ndg.xacml.core.expression.Expression derived type
        @return: list of attribute values for subject corresponding to given
        policy designator
        @rtype: list
        """
        return []



"""NDG Security Resource type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: resource.py 7365 2010-08-25 15:41:20Z pjkersha $"
from ndg.xacml.core import TargetChildBase
from ndg.xacml.core.match import ResourceMatch


class Resource(TargetChildBase):
    """XACML Resource Target Policy element
    
    @cvar MATCH_TYPE: Sets the type for match attributes in this 
    TargetChildBase derived class
    @type MATCH_TYPE: ndg.xacml.core.match.ResourceMatch
    @cvar ELEMENT_LOCAL_NAME: XML element local name
    @type ELEMENT_LOCAL_NAME: string
    """
    MATCH_TYPE = ResourceMatch
    ELEMENT_LOCAL_NAME = 'Resource'
    ID = "urn:oasis:names:tc:xacml:1.0:resource:resource-id"
    __slots__ = ()
    
    @property
    def resourceMatches(self):
        """Convenience wrapper to base class method
        @return: list of matches
        @rtype: ndg.xacml.utils.TypedList 
        """
        return self.matches



"""NDG XACML ElementTree based reader for Expression type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "18/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: expressionreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from abc import abstractmethod

from ndg.xacml.core.expression import Expression
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader


class ExpressionReader(ETreeAbstractReader):
    '''ElementTree based XACML Expression type parser
    
    @cvar TYPE: XACML type to instantiate from parsed object
    @type TYPE: abc.ABCMeta
    '''
    TYPE = Expression
    
    def __call__(self, obj):
        """Parse Expression object
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML expression instance
        @rtype: ndg.xacml.core.expression.Expression derived type 
        @raise XMLParseError: error reading element               
        """
        elem = super(ExpressionReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        expression = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.DATA_TYPE_ATTRIB_NAME,):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" element' %
                                    (attributeName, 
                                     xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue)
             
        expression.dataType, = attributeValues
                   
        self._parseExtension(elem, expression)
        
        return expression
        
    @abstractmethod
    def _parseExtension(self, elem, expression):
        """Derived classes should implement this method to read any remaining
        attributes and elements specific to their type
        
        @param elem: ElementTree XML element
        @type elem: xml.etree.Element
        
        @param expression: attribute selector to be updated with parsed
        values
        @type expression: ndg.xacml.core.attributevalue.AttributeValue
        
        
        @raise NotImplementedError: Derived classes should implement
        """
        raise NotImplementedError()
          
# Set up new class as an abstract base itself             
ETreeAbstractReader.register(ExpressionReader)



#!/usr/bin/env python
"""NDG XACML Context unit test package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: test_context.py 7443 2010-09-03 13:42:43Z pjkersha $"
import unittest
from os import path
import logging
logging.basicConfig(level=logging.DEBUG)

from ndg.xacml.test import XACML_NDGTEST1_FILEPATH
from ndg.xacml.parsers.etree.factory import ReaderFactory
from ndg.xacml.core import Identifiers
from ndg.xacml.core.context.pdpinterface import PDPInterface
from ndg.xacml.core.context.pdp import PDP
from ndg.xacml.core.context.handler import CtxHandlerInterface
from ndg.xacml.core.attribute import Attribute
from ndg.xacml.core.attributevalue import (AttributeValue, 
                                           AttributeValueClassFactory)
from ndg.xacml.core.context.request import Request
from ndg.xacml.core.context.response import Response
from ndg.xacml.core.context.result import Result, Decision
from ndg.xacml.core.context.subject import Subject
from ndg.xacml.core.context.resource import Resource
from ndg.xacml.core.context.action import Action

attributeValueFactory = AttributeValueClassFactory()
AnyUriAttributeValue = attributeValueFactory(AttributeValue.ANY_TYPE_URI)
StringAttributeValue = attributeValueFactory(AttributeValue.STRING_TYPE_URI)

ROLE_ATTRIBUTE_ID = "urn:ndg:security:authz:1.0:attr"
SUBJECT_ID = 'https://my.name.somewhere.ac.uk'

class TestContextHandler(CtxHandlerInterface):
    """Test implementation of Context Handler which includes an implemented PIP
    interface"""
    
    def __init__(self):
        """Add an attribute to hold a reference to a policy information point"""
        
        super(TestContextHandler, self).__init__()
        self.pip = None        
        
    def handlePEPRequest(self, myRequest):
        """Handle request from Policy Enforcement Point
        
        @param pepRequest: request from PEP, derived class determines its type
        e.g. SAML AuthzDecisionQuery
        @type myRequest: type
        @return: PEP response - derived class determines type
        @rtype: None
        """
        
        # Convert myRequest to XACML context request - var assignment here is 
        # representative of this process rather than actually doing anything.
        xacmlRequest = myRequest
        
        if self.pdp is None:
            raise TypeError('No "pdp" attribute set')
        
        # Add a reference to this context so that the PDP can invoke queries
        # back to the PIP
        xacmlRequest.ctxHandler = self 
               
        xacmlResponse = self.pdp.evaluate(xacmlRequest)
        
        # Convert XACML context response to domain specific request
        myResponse = xacmlResponse
        
        return myResponse
    
    def pipQuery(self, request, designator):
        '''PIP adds admin attribute value for given attribute ID and for any 
        subject'''
        if designator.attributeId == ROLE_ATTRIBUTE_ID:
            attrVal = StringAttributeValue(value='admin')
            return [attrVal]
        else:
            return None


class XacmlContextBaseTestCase(unittest.TestCase):
    """Base class containing common methods for test initialisation"""
    
    def _createRequestCtx(self, 
                          resourceId, 
                          includeSubject=True,
                          subjectRoles=('staff',),
                          roleAttributeId=ROLE_ATTRIBUTE_ID,
                          action='read'):
        """Create an example XACML Request Context for tests"""
        request = Request()
        
        if includeSubject:
            subject = Subject()
            openidSubjectAttribute = Attribute()
            
            
            openidSubjectAttribute.attributeId = "urn:esg:openid"
            openidSubjectAttribute.dataType = AnyUriAttributeValue.IDENTIFIER
            
            openidSubjectAttribute.attributeValues.append(
                                                        AnyUriAttributeValue())
            openidSubjectAttribute.attributeValues[-1].value = SUBJECT_ID
                                        
            
            subject.attributes.append(openidSubjectAttribute)
    
            for role in subjectRoles:
                roleAttribute = Attribute()
                
                roleAttribute.attributeId = roleAttributeId
                roleAttribute.dataType = StringAttributeValue.IDENTIFIER
                
                roleAttribute.attributeValues.append(StringAttributeValue())
                roleAttribute.attributeValues[-1].value = role 
            
                subject.attributes.append(roleAttribute)
                                      
            request.subjects.append(subject)
        
        resource = Resource()
        resourceAttribute = Attribute()
        resource.attributes.append(resourceAttribute)
        
        resourceAttribute.attributeId = Identifiers.Resource.RESOURCE_ID
                            
        resourceAttribute.dataType = AnyUriAttributeValue.IDENTIFIER
        resourceAttribute.attributeValues.append(AnyUriAttributeValue())
        resourceAttribute.attributeValues[-1].value = resourceId

        request.resources.append(resource)
        
        request.action = Action()
        actionAttribute = Attribute()
        request.action.attributes.append(actionAttribute)
        
        actionAttribute.attributeId = Identifiers.Action.ACTION_ID
        actionAttribute.dataType = StringAttributeValue.IDENTIFIER
        actionAttribute.attributeValues.append(StringAttributeValue())
        actionAttribute.attributeValues[-1].value = action
        
        return request
        
    def _createPDPfromPolicy(self):
        pdp = PDP.fromPolicySource(XACML_NDGTEST1_FILEPATH, ReaderFactory)
        return pdp    


class XacmlContextTestCase(XacmlContextBaseTestCase):
    """Test PDP, PAP, PIP and Context handler"""
    
    def test01CreateRequest(self):
        requestCtx = self._createRequestCtx("http://localhost")
        self.assert_(requestCtx)
        
    def test02CreateResponse(self):
        response = Response()
        result = Result()
        response.results.append(result)
        result.decision = Decision()
        result.decision.value = Decision.NOT_APPLICABLE
        
    def test03AbstractCtxHandler(self):
        self.assertRaises(TypeError, CtxHandlerInterface, 
                          "Context handler is an abstract base class")
        
    def test04CreateCtxHandler(self):
        ctxHandler = TestContextHandler()
        
    def test05PDPInterface(self):
        self.assertRaises(TypeError, PDPInterface)
        
    def test06CreatePDP(self):
        pdp = PDP()
        self.assert_(pdp)
        
    def test07CreatePDPfromPolicy(self):
        pdp = self._createPDPfromPolicy()
        self.assert_(pdp)
    
    
class XacmlEvalPdpWithPermitOverridesPolicy(XacmlContextBaseTestCase):
    """Test PDP with permit overrides rule combining algorithm"""
    
    NOT_APPLICABLE_RESOURCE_ID = 'https://localhost'
    
    # This could be any applicable resource value, provided there's no rule to
    # override and enable access
    PRIVATE_RESOURCE_ID = 'http://localhost/private-resource'
    
    PUBLIC_RESOURCE_ID = 'http://localhost/resource-only-restricted'
    NOT_APPLICABLE_RESOURCE_ID = 'https://localhost'
        
    SINGLE_SUBJECT_ROLE_RESTRICTED_ID = \
        'http://localhost/single-subject-role-restricted'
    ACTION_AND_SINGLE_SUBJECT_ROLE_RESTRICTED_ID = \
        'http://localhost/action-and-single-subject-role-restricted'
    AT_LEAST_ONE_SUBJECT_ROLE_RESTRICTED_ID = \
        'http://localhost/at-least-of-subject-role-restricted'
        
    def setUp(self):
        self.pdp = self._createPDPfromPolicy()
        
    def test01NotApplicable(self):
        # Set a resource Id that doesn't match the main target
        request = self._createRequestCtx(
                                    self.__class__.NOT_APPLICABLE_RESOURCE_ID)
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.NOT_APPLICABLE, 
                        "Expecting not applicable decision")
        
    def test02PublicallyAccessibleResource(self):
        # Test a resource which has no subject restrictions
        request = self._createRequestCtx(self.__class__.PUBLIC_RESOURCE_ID,
                                         includeSubject=False)
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.PERMIT, 
                        "Expecting Permit decision")
        
    def test03PrivateResource(self):
        request = self._createRequestCtx(
                                    self.__class__.PRIVATE_RESOURCE_ID)
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.DENY, 
                        "Expecting Deny decision")

    def test04SingleSubjectRoleRestrictedResource(self):
        # Access based on a resource ID and single subject role
        request = self._createRequestCtx(
                            self.__class__.SINGLE_SUBJECT_ROLE_RESTRICTED_ID)
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.PERMIT, 
                        "Expecting Permit decision")  

    def test05SingleSubjectRoleRestrictedResourceDeniesAccess(self):
        # Subject doesn't have the required role for access
        request = self._createRequestCtx(
                            self.__class__.SINGLE_SUBJECT_ROLE_RESTRICTED_ID,
                            subjectRoles=('student',))
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.DENY, 
                        "Expecting Deny decision")  

    def test06ActionAndSingleSubjectRoleRestrictedResource(self):
        # Test restriction based on action type as well as subject role
        request = self._createRequestCtx(
                    self.__class__.ACTION_AND_SINGLE_SUBJECT_ROLE_RESTRICTED_ID)
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.PERMIT, 
                        "Expecting Permit decision")

    def test07ActionAndSingleSubjectRoleRestrictedResourceDeniesAccess(self):
        # Test subject requests invalid action type
        request = self._createRequestCtx(
                    self.__class__.ACTION_AND_SINGLE_SUBJECT_ROLE_RESTRICTED_ID,
                    action='write')
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.DENY, 
                        "Expecting Deny decision")  

    def test08AtLeastOneSubjectRoleResource(self):
        # Test at least one member function
        request = self._createRequestCtx(
                    self.__class__.AT_LEAST_ONE_SUBJECT_ROLE_RESTRICTED_ID,
                    action='write')
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.PERMIT, 
                        "Expecting Permit decision")             

    def test09AtLeastOneSubjectRoleResourceDeniesAccess(self):
        # Test at least one member function where subject doesn't have one of
        # the required roles
        request = self._createRequestCtx(
                    self.__class__.AT_LEAST_ONE_SUBJECT_ROLE_RESTRICTED_ID,
                    subjectRoles=('student',))
        response = self.pdp.evaluate(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.DENY, 
                        "Expecting Deny decision")             
    
    def test10PipAddsRequiredAttributeValToEnableAccess(self):
        # The PDP is part of a context handler with a PIP which adds subject
        # attributes under prescribed conditions on the evaluation of 
        # subject attribute designators.  In this case the addition of the PIP
        # adds an attribute value to one of the subject's attributes which means
        # they're granted access where otherwise access would be denied
        ctxHandler = TestContextHandler()
        ctxHandler.pdp = self.pdp
        
        request = self._createRequestCtx(
                    self.__class__.AT_LEAST_ONE_SUBJECT_ROLE_RESTRICTED_ID,
                    subjectRoles=('student',))
        
        response = ctxHandler.handlePEPRequest(request)
        self.failIf(response is None, "Null response")
        for result in response.results:
            self.failIf(result.decision != Decision.PERMIT, 
                        "Expecting PERMIT decision")         
        
                                
if __name__ == "__main__":
    unittest.main()


"""NDG Security Policy type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: policy.py 7351 2010-08-20 14:46:35Z pjkersha $"
import traceback
import logging
log = logging.getLogger(__name__)

from ndg.xacml.utils import TypedList
from ndg.xacml.parsers import AbstractReaderFactory, AbstractReader
from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.policydefaults import PolicyDefaults
from ndg.xacml.core.target import Target
from ndg.xacml.core.rule import Rule
from ndg.xacml.core.obligation import Obligation
from ndg.xacml.core.rule_combining_alg import (RuleCombiningAlgClassFactory,
                                               RuleCombiningAlgInterface)
from ndg.xacml.core.functions import (UnsupportedStdFunctionError,
                                      UnsupportedFunctionError)

# Evaluation of a request context
from ndg.xacml.core.context.response import Response
from ndg.xacml.core.context.result import Result, Decision, StatusCode
from ndg.xacml.core.context.exceptions import XacmlContextError


class PolicyParseError(Exception):
    """Error reading policy attributes from file"""


class InvalidPolicyXmlNsError(PolicyParseError):
    """Invalid XML namespace for policy document"""


class Policy(XacmlCoreBase):
    """XACML Policy
    
    @cvar DEFAULT_XACML_VERSION: default is 2.0
    @type DEFAULT_XACML_VERSION: string
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar POLICY_ID_ATTRIB_NAME: policy id XML attribute name
    @type POLICY_ID_ATTRIB_NAME: string
    @cvar RULE_COMBINING_ALG_ID_ATTRIB_NAME: rule combining algorithm id XML
    attribute name
    @type RULE_COMBINING_ALG_ID_ATTRIB_NAME: string
    @cvar VERSION_ATTRIB_NAME: version XML attribute name
    @type VERSION_ATTRIB_NAME: string
    @cvar DESCRIPTION_LOCAL_NAME: description XML element local name
    @type DESCRIPTION_LOCAL_NAME: string
    @cvar POLICY_DEFAULTS_LOCAL_NAME: policy defaults XML element local name
    @type POLICY_DEFAULTS_LOCAL_NAME: string
    @cvar COMBINER_PARAMETERS_LOCAL_NAME: combiner parameter XML element local
    name
    @type COMBINER_PARAMETERS_LOCAL_NAME: string
    @cvar RULE_COMBINER_PARAMETERS_LOCAL_NAME: rule combiner parameter XML
    element local name
    @type RULE_COMBINER_PARAMETERS_LOCAL_NAME: string
    @cvar OBLIGATIONS_LOCAL_NAME: obligations XML element local name
    @type OBLIGATIONS_LOCAL_NAME: string
    
    @ivar __policyId: policy id
    @type __policyId: NoneType / basestring
    @ivar __version: policy version
    @type __version: NoneType / basestring
    @ivar __ruleCombiningAlgId: rule combining algorithm ID
    @type __ruleCombiningAlgId: NoneType / basestring
    @ivar __description: policy decription text
    @type __description: NoneType / basestring
    @ivar __target: target element
    @type __target: NoneType / ndg.xacml.core.target.Target
    @ivar __attr: list of rules
    @type __attr: ndg.xacml.utils.TypedList
    @ivar __obligations: obligations
    @type __obligations: ndg.xacml.utils.TypedList
    @ivar __ruleCombiningAlgFactory: rule combining algorithm factory
    @type __ruleCombiningAlgFactory: ndg.xacml.core.rule_combining_alg.RuleCombiningAlgClassFactory
    @ivar __ruleCombiningAlg: rule combining algorithm
    @type __ruleCombiningAlg: NoneType / ndg.xacml.core.rule_combining_alg.RuleCombiningAlgInterface
    """ 
    DEFAULT_XACML_VERSION = "2.0"
    ELEMENT_LOCAL_NAME = "Policy"
    POLICY_ID_ATTRIB_NAME = "PolicyId"
    RULE_COMBINING_ALG_ID_ATTRIB_NAME = "RuleCombiningAlgId"
    VERSION_ATTRIB_NAME = "Version"

    DESCRIPTION_LOCAL_NAME = "Description"
    POLICY_DEFAULTS_LOCAL_NAME = "PolicyDefaults"
    COMBINER_PARAMETERS_LOCAL_NAME = "CombinerParameters"
    RULE_COMBINER_PARAMETERS_LOCAL_NAME = "RuleCombinerParameters"
    OBLIGATIONS_LOCAL_NAME = "Obligations"
    
    __slots__ = (
        '__policyId',
        '__version',
        '__ruleCombiningAlgId',
        '__description',
        '__policyDefaults',
        '__target',
        '__attr',
        '__obligations',
        '__ruleCombiningAlgFactory',
        '__ruleCombiningAlg'
    )
    
    def __init__(self, ruleCombiningAlgFactory=None):
        """Customise rule combining behaviour by passing in a custom combining
        algorithm factory.  This is invoked when the combining algorithm Id
        property is set in order to create the corresponding combining algorithm
        object
        
        @param ruleCombiningAlgFactory: factory object for return a rule 
        combining algorithm class for a given URI.  Defaults to 
        @type ruleCombiningAlgFactory: NoneType / defaults to 
        ndg.xacml.core.rule_combining_alg.RuleCombiningAlgClassFactory
        """
        super(Policy, self).__init__()
        self.__policyId = None
        self.__version = None
        self.__ruleCombiningAlgId = None
        self.__description = None
        self.__target = None
        self.__policyDefaults = None
        
        # Attr should eventually allow a choice of Rule, CombinerParameter, 
        # RuleCombinerParameter and VariableDefinition but only Rule type is 
        # currently supported
        self.__attr = TypedList(Rule)
        
        self.__obligations = TypedList(Obligation)
        
        self.__ruleCombiningAlgFactory = None
        if ruleCombiningAlgFactory is None:
            self.ruleCombiningAlgFactory = RuleCombiningAlgClassFactory()
        else:
            self.ruleCombiningAlgFactory = ruleCombiningAlgFactory

        self.__ruleCombiningAlg = None

    def _getRuleCombiningAlgFactory(self):
        """
        @return: rule combining algorithm factory
        @rtype: NoneType / ndg.xacml.core.rule_combining_alg.RuleCombiningAlgClassFactory
        """
        return self.__ruleCombiningAlgFactory

    def _setRuleCombiningAlgFactory(self, value):
        """
        @param value: rule combining algorithm factory
        @type value: ndg.xacml.core.rule_combining_alg.RuleCombiningAlgClassFactory
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, RuleCombiningAlgClassFactory):
            raise TypeError('Expecting %r derived type for '
                            '"ruleCombiningAlgFactory" attibute; got %r' % 
                            (RuleCombiningAlgClassFactory, type(value)))
            
        self.__ruleCombiningAlgFactory = value

    ruleCombiningAlgFactory = property(_getRuleCombiningAlgFactory, 
                                       _setRuleCombiningAlgFactory, 
                                       doc="Rule Combining Algorithm Factory")
    
    @property
    def ruleCombiningAlg(self):
        """Rule Combining algorithm
        @return: rule combining algorithm class instance
        @rtype: ndg.xacml.core.rule_combining_alg.RuleCombiningAlgInterface 
        derived type
        """
        return self.__ruleCombiningAlg
    
    @classmethod
    def fromSource(cls, source, readerFactory):
        """Create a new policy from the input source parsing it using a 
        reader from the required reader factory e.g. ETreeReaderFactory to use
        ElementTree based parsing
        
        @param source: source from which to read the policy - file path,
        file object, XML node or other dependent on the reader factory selected
        @type source: string, file, XML node type
        @param readerFactory: factory class returns reader class used to parse 
        the policy
        @type readerFactory: ndg.xacml.parsers.AbstractReaderFactory
        @return: new policy instance
        @rtype: ndg.xacml.core.policy.Policy
        """
        if not issubclass(readerFactory, AbstractReaderFactory):
            raise TypeError('Expecting %r derived class for reader factory '
                            'method; got %r' % (AbstractReaderFactory, 
                                                readerFactory))
            
        reader = readerFactory.getReader(cls)
        if not issubclass(reader, AbstractReader):
            raise TypeError('Expecting %r derived class for reader class; '
                            'got %r' % (AbstractReader, reader))
            
        return reader.parse(source)
        
    def _getPolicyId(self):
        '''
        @return: policy id
        @rtype: NoneType / basestring
        '''
        return self.__policyId

    def _setPolicyId(self, value):
        '''@param value: policy id
        @type value: basestring
        @raise TypeError: incorrect input type
        '''
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "policyId" '
                            'attribute; got %r' % type(value))
            
        self.__policyId = value

    policyId = property(_getPolicyId, _setPolicyId, None, "Policy Id")

    def _getVersion(self):
        '''@return: policy version
        @rtype: NoneType / basestring
        '''
        return self.__version

    def _setVersion(self, value):
        '''@param value: policy version
        @type value: basestring
        @raise TypeError: incorrect input type
        '''
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "version" '
                            'attribute; got %r' % type(value))
            
        self.__version = value

    version = property(_getVersion, _setVersion, None, "Policy Version")

    def _getRuleCombiningAlgId(self):
        '''@return: rule combining algorithm ID
        @rtype: NoneType / basestring
        '''
        return self.__ruleCombiningAlgId

    def _setRuleCombiningAlgId(self, value):
        '''@param value: rule combining algorithm ID
        @type value: NoneType / basestring
        @raise TypeError: incorrect input type
        '''
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "ruleCombiningAlgId" '
                            'attribute; got %r' % type(value))
            
        self.__ruleCombiningAlgId = value
        self._setRuleCombiningAlgFromId()
        
    def _setRuleCombiningAlgFromId(self):
        """Set the rule combining algorithm implementation from the Id set in
        __ruleCombiningAlgId the attribute
        
        @raise TypeError: incorrect input type
        @raise UnsupportedStdFunctionError: no implementation is avaliable for
        this XACML rule combining algorithm 
        @raise UnsupportedFunctionError: the rule combining algorithm is not
        recognised as a standard XACML one
        """
        # Look-up rule combining algorithm
        ruleCombiningAlgClass = self.__ruleCombiningAlgFactory(
                                                    self.__ruleCombiningAlgId)
        if (not isinstance(ruleCombiningAlgClass, type) or 
            not issubclass(ruleCombiningAlgClass, RuleCombiningAlgInterface)):
            raise TypeError('Expecting %r derived type for rule combining '
                            'algorithm class; got %r type' %
                            (RuleCombiningAlgInterface, ruleCombiningAlgClass))
            
        self.__ruleCombiningAlg = ruleCombiningAlgClass()
        if self.__ruleCombiningAlg is NotImplemented:
            raise UnsupportedStdFunctionError('The rule combining algorithm %r '
                                              'is not currently implemented' % 
                                              self.__ruleCombiningAlgId)
            
        elif self.__ruleCombiningAlg is None:
            raise UnsupportedFunctionError('%r is not recognised as a valid '
                                           'XACML rule combining algorithm' % 
                                           self.__ruleCombiningAlgId) 

    ruleCombiningAlgId = property(_getRuleCombiningAlgId, 
                                  _setRuleCombiningAlgId, None, 
                                  doc="Rule Combining Algorithm Id")

    @property
    def combinerParameters(self):
        """@raise NotImplementedError: combiner parameters property is not
        currently implemented
        """
        raise NotImplementedError()
    
    @property
    def ruleCombinerParameters(self):
        """@raise NotImplementedError: rule combiner parameters property is not
        currently implemented
        """
        raise NotImplementedError()
    
    @property
    def variableDefinitions(self):
        """@raise NotImplementedError: variable definitions parameters property 
        is not currently implemented
        """
        raise NotImplementedError()
    
    @property
    def rules(self):
        """Return the list of rules
        @return: list of rules
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__attr
    
    @property
    def obligations(self):
        """@return: obligations
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__obligations 

    def _getTarget(self):
        """@return: target element
        @rtype: NoneType / ndg.xacml.core.target.Target
        """
        return self.__target

    def _setTarget(self, value):
        """@param value: target element
        @type value: ndg.xacml.core.target.Target
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, Target):
            raise TypeError('Expecting Target for "target" '
                            'attribute; got %r' % type(value))
        self.__target = value

    target = property(_getTarget, _setTarget, doc="list of Policy targets")

    def _getDescription(self):
        '''@return: policy description text
        @rtype: NoneType / basestring
        '''
        return self.__description

    def _setDescription(self, value):
        '''@param value: policy description text
        @type value: basestring
        @raise TypeError: incorrect input type
        '''
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "description" '
                            'attribute; got %r' % type(value))
        self.__description = value

    description = property(_getDescription, _setDescription, 
                           doc="Policy Description text")

    def _getPolicyDefaults(self):
        '''@return: policy defaults
        @rtype: NoneType / ndg.xacml.core.policydefaults.PolicyDefaults
        '''
        return self.__policyDefaults

    def _setPolicyDefaults(self, value):
        '''@param value: policy defaults
        @type value: ndg.xacml.core.policydefaults.PolicyDefaults
        @raise TypeError: incorrect input type
        '''
        if not isinstance(value, PolicyDefaults):
            raise TypeError('Expecting string type for "policyDefaults" '
                            'attribute; got %r' % type(value))
            
        self.__policyDefaults = value

    policyDefaults = property(_getPolicyDefaults, 
                              _setPolicyDefaults, 
                              None, 
                              "Policy PolicyDefaults element")   

    def evaluate(self, request):
        """Make an access control decision for the given request based on this
        policy
        
        @param request: XACML request context
        @type request: ndg.xacml.core.context.request.Request
        @return: XACML response instance
        @rtype: ndg.xacml.core.context.response.Response
        @raise XacmlContextError: error evaluating input request context
        """
        response = Response()
        result = Result.createInitialised(decision=Decision.NOT_APPLICABLE)
        response.results.append(result)
            
        # Exception block around all rule processing in order to set
        # INDETERMINATE response from any exceptions raised
        try:
            log.debug('Checking policy %r target for match with request...',
                      self.policyId)
            
            target = self.target
            
            # If no target is set, ALL requests are counted as matches
            if target is not None:
                if not target.match(request):
                    # The target didn't match so the whole policy does not apply
                    # to this request
                    log.debug('No match for policy target setting %r decision',
                              Decision.NOT_APPLICABLE_STR)
                    
                    result.decision = Decision.NOT_APPLICABLE
                    return response
            
            log.debug('Request matches the Policy %r target', self.policyId)
            
            # Apply the rule combining algorithm here combining the 
            # effects from the rules evaluated into an overall decision
            result.decision = self.ruleCombiningAlg.evaluate(self.rules,
                                                             request)
        except XacmlContextError, e:
            log.error('Exception raised evaluating request context, returning '
                      '%r decision:%s', 
                      e.response.results[0].decision, 
                      traceback.format_exc())
            
            result = e.response.results[0]
            
        except Exception:
            # Catch all so that nothing is handled from within the scope of this
            # method
            log.error('No PDPError type exception raised evaluating request '
                      'context, returning %r decision:%s', 
                      Decision.INDETERMINATE_STR, 
                      traceback.format_exc()) 
                       
            result.decision = Decision.INDETERMINATE
            result.status.statusCode.value = StatusCode.PROCESSING_ERROR
            
        return response




'''
Created on 24 Feb 2010

@author: pjkersha
'''
from ndg.xacml.utils import TypedList
"""NDG Security Target type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: target.py 7260 2010-08-02 12:28:59Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.action import Action
from ndg.xacml.core.resource import Resource
from ndg.xacml.core.subject import Subject
from ndg.xacml.core.environment import Environment


class Target(XacmlCoreBase):
    """XACML Target element
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar SUBJECTS_ELEMENT_LOCAL_NAME: XML local name for the subjects element
    @type SUBJECTS_ELEMENT_LOCAL_NAME: string
    @cvar ACTIONS_ELEMENT_LOCAL_NAME: XML local name for the actions element
    @type ACTIONS_ELEMENT_LOCAL_NAME: string
    @cvar RESOURCES_ELEMENT_LOCAL_NAME: XML local name for the resources element
    @type RESOURCES_ELEMENT_LOCAL_NAME: string
    @cvar ENVIRONMENTS_ELEMENT_LOCAL_NAME: XML local name for the environments
    element
    @type ENVIRONMENTS_ELEMENT_LOCAL_NAME: string
    @cvar CHILD_ATTRS: list of the XML child element names for <Target/>
    @type CHILD_ATTRS: tuple
    
    @ivar __subjects: list of subjects for this target
    @type __subjects: ndg.xacml.utils.TypedList
    @ivar __resources: list of resources for this target
    @type __resources: ndg.xacml.utils.TypedList
    @ivar __actions: list of actions for this target
    @type __actions: ndg.xacml.utils.TypedList
    @ivar __environments: list of environment settings for this target
    @type __environments: ndg.xacml.utils.TypedList
    """
    ELEMENT_LOCAL_NAME = "Target"
    SUBJECTS_ELEMENT_LOCAL_NAME = "Subjects"
    ACTIONS_ELEMENT_LOCAL_NAME = "Actions"
    RESOURCES_ELEMENT_LOCAL_NAME = "Resources"
    ENVIRONMENTS_ELEMENT_LOCAL_NAME = "Environments"
    CHILD_ATTRS = ('subjects', 'resources', 'actions', 'environments')
   
    __slots__ = ('__subjects', '__resources', '__actions', '__environments')
    
    def __init__(self):
        """Initial attributes"""
        self.__subjects = TypedList(Subject)
        self.__resources = TypedList(Resource)
        self.__actions = TypedList(Action)
        self.__environments = TypedList(Environment)
    
    @property
    def subjects(self):
        """Get subjects
        @return: list of subjects for this target
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__subjects
    
    @property
    def resources(self):
        """Get resources
        @return: list of resources for this target
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__resources
    
    @property
    def actions(self):
        """Get actions
        @return: list of actions for this target
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__actions
    
    @property
    def environments(self):
        """Get environments
        @return: list of environments for this target
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__environments
            
    def match(self, request):
        """Generic method to match a <Target> element to the request context
        
        @param request: XACML request context
        @type request: ndg.xacml.core.context.request.Request
        @return: True if request context matches the given target, 
        False otherwise
        @rtype: bool
        """
        
        # From section 5.5 of the XACML 2.0 Core Spec:
        #
        # For the parent of the <Target> element to be applicable to the 
        # decision request, there MUST be at least one positive match between 
        # each section of the <Target> element and the corresponding section of 
        # the <xacml-context:Request> element.
        # 
        # Also, 7.6:
        #
        # The target value SHALL be "Match" if the subjects, resources, actions 
        # and environments specified in the target all match values in the 
        # request context. 
        statusValues = [False]*len(self.__class__.CHILD_ATTRS) 
        
        # Iterate for target subjects, resources, actions and environments 
        # elements
        for i, attrName in enumerate(self.__class__.CHILD_ATTRS):
            # If any one of the <Target> children is missing then it counts as
            # a match e.g. for <Subjects> child element - Section 5.5:
            #
            # <Subjects> [Optional] Matching specification for the subject 
            # attributes in the context. If this element is missing,
            # then the target SHALL match all subjects.
            targetElem = getattr(self, attrName)
            if len(targetElem) == 0:
                statusValues[i] = True
                continue
            
            # Iterate over each for example, subject in the list of subjects:
            # <Target>
            #     <Subjects>
            #          <Subject>
            #              ...
            #          </Subject>
            #          <Subject>
            #              ...
            #          </Subject>
            #     ...
            # or resource in the list of resources and so on
            for targetSubElem in targetElem:
                if self._matchChild(targetSubElem, request):
                    # Within the list of e.g. subjects if one subject 
                    # matches then this counts as a subject match overall 
                    # for this target
                    statusValues[i] = True
 
        # Target matches if all the children (i.e. subjects, resources, actions
        # and environment sections) have at least one match.  Otherwise it 
        # doesn't count as a match
        return all(statusValues)
    
    def _matchChild(self, targetChild, request):
        """Match a request child element (a <Subject>, <Resource>, <Action> or 
        <Environment>) with the corresponding target's <Subject>, <Resource>, 
        <Action> or <Environment>.
        
        @param targetChild: Target Subject, Resource, Action or Environment
        object
        @type targetChild: ndg.xacml.core.TargetChildBase
        @param request: Request context object
        @type request: ndg.xacml.core.context.request.Request
        @return: True if request context matches something in the target
        @rtype: bool
        @raise UnsupportedStdFunctionError: policy references a function type 
        which is in the XACML spec. but is not supported by this implementation
        @raise UnsupportedFunctionError: policy references a function type which
        is not supported by this implementation
        """
        if targetChild is None:
            # Default if target child is not set is to match all children
            return True
        
        matchStatusValues = [True]*len(targetChild.matches)
        
        # Section 7.6
        #
        # A subject, resource, action or environment SHALL match a value in the
        # request context if the value of all its <SubjectMatch>, 
        # <ResourceMatch>, <ActionMatch> or <EnvironmentMatch> elements, 
        # respectively, are "True".
        #
        # e.g. for <SubjectMatch>es in <Subject> ...
        for i, childMatch in enumerate(targetChild.matches):
            matchStatusValues[i] = childMatch.evaluate(request)
            
        # Any match => overall match      
        return any(matchStatusValues)




"""NDG XACML core package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7365 2010-08-25 15:41:20Z pjkersha $"
from ndg.xacml.utils import TypedList

    
class XacmlCoreBase(object):
    """Base class for all XACML types
    
    @cvar XACML_1_0_NS_PREFIX: XACML version 1.0 namespace prefix
    @type XACML_1_0_NS_PREFIX: string
    @cvar XACML_2_0_NS_PREFIX: XACML version 2.0 namespace prefix
    @type XACML_2_0_NS_PREFIX: string
    @cvar XMLNS: list of valid XACML namespaces
    @type XMLNS: tuple
    @cvar ELEMENT_LOCAL_NAME: XML element local name for the given type
    @type ELEMENT_LOCAL_NAME: NoneType but implement as string in derived 
    classes
    
    @ivar __xmlns: XML namespace for the XACML type
    @type __xmlns: NoneType / basestring
        
    @ivar __elem: XML element 
    @type __elem: NoneType / dependent on Python XML parser used
    """
    XACML_1_0_NS_PREFIX = "urn:oasis:names:tc:xacml:1.0"
    XACML_2_0_NS_PREFIX = "urn:oasis:names:tc:xacml:2.0"
    
    XMLNS = (XACML_1_0_NS_PREFIX, XACML_2_0_NS_PREFIX)
    
    __slots__ = ('__xmlns', '__reader', '__writer', '__elem')

    ELEMENT_LOCAL_NAME = None
    
    def __init__(self):
        """Element local name check makes this a virtual method
        
        @raise NotImplementedError: derived classes must set 
        ELEMENT_LOCAL_NAME to a string
        """
        self.__xmlns = None
        self.__elem = None
        
        if not isinstance(self.__class__.ELEMENT_LOCAL_NAME, basestring):
            raise NotImplementedError('"ELEMENT_LOCAL_NAME" must be defined in '
                                      'a derived class')
        
    def _getXmlns(self):
        """Get XML Namespace for this XACML type
        @return: the XML namespace set
        @rtype: basestring/NoneType        
        """
        return self.__xmlns

    def _setXmlns(self, value):
        """Set XML Namespace for this XACML type
        @param value: the XML namespace to set
        @type value: basestring/NoneType                 
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting string type for "xmlns" '
                            'attribute; got %r' % type(value))
        self.__xmlns = value

    xmlns = property(_getXmlns, _setXmlns, 
                     doc="XML Namespace for policy the document")
    
    @property
    def isValidXmlns(self):
        """Check XML namespace fits with the known XACML namespaces
        @return: True if valid, False otherwise
        @rtype: bool
        """
        return self.xmlns in XacmlCoreBase.XMLNS
        
    @property
    def elem(self):
        """XML Node for as represented by parser/writer specified with the 
        reader/writer attributes.  Readers of context elements should set this
        element if a policy uses AttributeSelectors to do XPath queries into
        the request context
        """
        return self.__elem
        
    @elem.setter
    def elem(self, value):
        """"XML Node for as represented by parser/writer specified with the 
        reader/writer attributes
        
        @param value: XML node instance
        @type value: type (governed by reader/writer set for this XACML object)
        """
        self.__elem = value
            

class XacmlPolicyBase(XacmlCoreBase):
    """Base class for policy types
    
    @cvar XACML_2_0_POLICY_NS: XACML 2.0 policy XML namespace
    @type XACML_2_0_POLICY_NS: string
    """
    XACML_2_0_POLICY_NS = (XacmlCoreBase.XACML_2_0_NS_PREFIX +
                           ":policy:schema:os")
    __slots__ = ()
    
    def __init__(self):
        """Initialise parent class xmlns attribute based on this classes'
        policy namespace
        """
        super(XacmlPolicyBase, self).__init__()
        self.xmlns = XacmlPolicyBase.XACML_2_0_POLICY_NS
            
        
class TargetChildBase(XacmlPolicyBase):
    """Abstract Base class for XACML Policy Subject, Resource, Action and 
    Environment types: e.g. ndg.xacml.core.subject.Subject
    
    @cvar MATCH_TYPE: Set the type for match attributes in the derived class
    implementation e.g. ResourceMatch, SubjectMatch etc.
    @type MATCH_TYPE: NoneType - derived class must implement
    
    @ivar __matches: list of matches for this target
    @type __matches: ndg.xacml.core.utils.TypedList
    """
    MATCH_TYPE = None
    
    __slots__ = ('__matches', )
    
    def __init__(self):
        super(TargetChildBase, self).__init__()
        
        # Derived types can specify the type for matches via the MATCH_TYPE
        # class variable
        if self.__class__.MATCH_TYPE is None:
            raise NotImplementedError('Match type attribute must be specified '
                                      'in a derived class')
        self.__matches = TypedList(self.__class__.MATCH_TYPE)
        
    @property
    def matches(self):
        """Get matches list for this target
        @return: list of matches
        @rtype: ndg.xacml.core.utils.TypedList
        """
        return self.__matches
    
    
XACML_1_0_PREFIX = "urn:oasis:names:tc:xacml:1.0:"

class Identifiers(object):
    """XACML Identifiers"""
    class Subject(object):
        """XAMCL Subject Identifiers"""
        AUTHN_LOCALITY_DNS_NAME = XACML_1_0_PREFIX + \
            "subject:authn-locality:dns-name"
        AUTHN_LOCALITY_IP_ADDRESS = XACML_1_0_PREFIX + \
            "subject:authn-locality:ip-address"
        AUTHN_METHOD = XACML_1_0_PREFIX + "subject:authentication-method"
        AUTHN_TIME = XACML_1_0_PREFIX + "subject:authentication-time"
        KEY_INFO = XACML_1_0_PREFIX + "subject:key-info"
        REQUEST_TIME = XACML_1_0_PREFIX + "subject:request-time"
        SESSION_START_TIME = XACML_1_0_PREFIX + "subject:session-start-time"
        SUBJECT_ID = XACML_1_0_PREFIX + "subject:subject-id"
        SUBJECT_ID_QUALIFIER = XACML_1_0_PREFIX + "subject:subject-id-qualifier"
        
    class SubjectCategory(object):
        """XACML Subject Category Identifiers"""
        ACCESS_SUBJECT = XACML_1_0_PREFIX + "subject-category:access-subject"
        CODEBASE = XACML_1_0_PREFIX + "subject-category:codebase"
        INTERMEDIARY_SUBJECT = XACML_1_0_PREFIX + \
            "subject-category:intermediary-subject"
        RECIPIENT_SUBJECT = XACML_1_0_PREFIX + \
            "subject-category:recipient-subject"
        REQUESTING_MACHINE = XACML_1_0_PREFIX + \
            "subject-category:requesting-machine"
        
    class Resource(object):
        """XACML Resource Identifiers"""
        RESOURCE_LOCATION = XACML_1_0_PREFIX + "resource:resource-location"
        RESOURCE_ID = XACML_1_0_PREFIX + "resource:resource-id"
        SIMPLE_FILE_NAME = XACML_1_0_PREFIX + "resource:simple-file-name"
        
    class Action(object):
        """XACML Action Identifiers"""
        ACTION_ID = XACML_1_0_PREFIX + "action:action-id"
        IMPLIED_ACTION = XACML_1_0_PREFIX + "action:implied-action"
     
    class Environment(object):
        """XACML Environment Identifiers"""
        CURRENT_TIME = XACML_1_0_PREFIX + "environment:current-time"
        CURRENT_DATE = XACML_1_0_PREFIX + "environment:current-date"
        CURRENT_DATETIME = XACML_1_0_PREFIX + "environment:current-dateTime"






"""NDG XACML ElementTree based Apply type reader 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "19/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: applyreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core.apply import Apply
from ndg.xacml.core.attributevalue import AttributeValue
from ndg.xacml.core.variablereference import VariableReference
from ndg.xacml.core.attributeselector import AttributeSelector
from ndg.xacml.core.attributedesignator import (SubjectAttributeDesignator,
                                                EnvironmentAttributeDesignator,
                                                ActionAttributeDesignator,
                                                EnvironmentAttributeDesignator)
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.expressionreader import ExpressionReader
from ndg.xacml.parsers.etree.factory import ReaderFactory


class ApplyReader(ETreeAbstractReader):
    '''ElementTree based XACML Apply type parser
    
    @cvar FUNCTION_ELEMENT_LOCAL_NAME: XML local name for function element
    @type FUNCTION_ELEMENT_LOCAL_NAME: string
    
    @cvar VARIABLE_REFERENCE_ELEMENT_LOCAL_NAME: XML local name for variable
    reference element
    @type VARIABLE_REFERENCE_ELEMENT_LOCAL_NAME: string
    
    @cvar TYPE: XACML class type that this reader will read values into
    @type TYPE: abc.ABCMeta
    '''
    TYPE = Apply
    
    # These two are not currently implemented.  When an implementation is made
    # the ELEMENT_LOCAL_NAME may be referenced from the native class rather than
    # a class variable here
    FUNCTION_ELEMENT_LOCAL_NAME = 'Function'
    VARIABLE_REFERENCE_ELEMENT_LOCAL_NAME = 'VariableReference'
    
    def __call__(self, obj):
        """Parse Apply type object

        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: ElementTree element
        @rtype: xml.etree.Element
        """
        elem = super(ApplyReader, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        applyObj = xacmlType()
        
        if QName.getLocalPart(elem.tag) != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
            
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.FUNCTION_ID_ATTRIB_NAME, ):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" '
                                    'element' % (attributeName,
                                                 xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue) 
                   
        applyObj.functionId, = attributeValues
        
        # Allow for any of the defined Expression sub-types in the child 
        # elements
        for subElem in elem:
            localName = QName.getLocalPart(subElem.tag)
            if localName == xacmlType.ELEMENT_LOCAL_NAME:
                applyObj.expressions.append(ApplyReader.parse(subElem))
             
            elif localName == AttributeValue.ELEMENT_LOCAL_NAME:
                AttributeValueReader = ReaderFactory.getReader(AttributeValue) 
                applyObj.expressions.append(AttributeValueReader.parse(subElem))
                 
            elif localName == SubjectAttributeDesignator.ELEMENT_LOCAL_NAME:
                SubjectAttributeDesignatorReader = ReaderFactory.getReader(
                                                SubjectAttributeDesignator)
                applyObj.expressions.append(
                                SubjectAttributeDesignatorReader.parse(subElem))
                
            elif localName == EnvironmentAttributeDesignator.ELEMENT_LOCAL_NAME:
                EnvironmentAttributeDesignatorReader = ReaderFactory.getReader(
                                                EnvironmentAttributeDesignator)
                applyObj.expressions.append(
                            EnvironmentAttributeDesignatorReader.parse(subElem))
                
            elif localName == ActionAttributeDesignator.ELEMENT_LOCAL_NAME:
                ActionAttributeDesignatorReader = ReaderFactory.getReader(
                                                ActionAttributeDesignator)
                applyObj.expressions.append(
                                ActionAttributeDesignatorReader.parse(subElem))
                
            elif localName == EnvironmentAttributeDesignator.ELEMENT_LOCAL_NAME:
                EnvironmentAttributeDesignatorReader = ReaderFactory.getReader(
                                                EnvironmentAttributeDesignator)
                applyObj.expressions.append(
                            EnvironmentAttributeDesignatorReader.parse(subElem))
                
            elif localName == AttributeSelector.ELEMENT_LOCAL_NAME:
                AttributeSelectorReader = ReaderFactory.getReader(
                                                            AttributeSelector)
                applyObj.expressions.append(
                                        AttributeSelectorReader.parse(subElem))
            
            elif localName == Condition.ELEMENT_LOCAL_NAME:
                ConditionReader = ReaderFactory.getReader(Condition)
                applyObj.expressions.append(ConditionReader.parse(subElem))
                
            elif localName == self.__class__.FUNCTION_ELEMENT_LOCAL_NAME:
                raise NotImplementedError('%r Apply sub-element not '
                                          'implemented', localName)
            
            elif (localName == VariableReference.ELEMENT_LOCAL_NAME):                                
                raise NotImplementedError('%r Apply sub-element not '
                                          'implemented', localName)
            else:
                raise XMLParseError('%r Apply sub-element not recognised', 
                                    localName)
   
        return applyObj


"""NDG Security Policy Decision Point interface definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: pdpinterface.py 7087 2010-06-25 11:23:09Z pjkersha $"
from abc import ABCMeta, abstractmethod
from ndg.xacml.core.context.request import Request


class PDPInterface(object):
    """Interface class for XACML Policy Enforcement Point"""
    __metaclass__ = ABCMeta
    __slots__ = ()
    
    @abstractmethod
    def evaluate(self, request):
        '''evaluate the input request and return an access control decision
        in the returned response
        
        @param request: XACML context request
        @type request: ndg.xacml.core.context.request.Request
        @return: XACML context response
        @rtype: None (this abstract method) expecting 
        ndg.xacml.core.context.response.Response type in implementations of this
        class
        '''
        if not isinstance(request, Request):
            raise TypeError('Expecting %r type for input request; got %r '
                            'instead' % (Request, type(request)))



"""NDG Security Policy Access Point type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: pap.py 7087 2010-06-25 11:23:09Z pjkersha $"



"""NDG XACML match function module - contains implementations for XACML 1.0
*-match functions

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = ""
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = '$Id: $'
from ndg.xacml.core.context.exceptions import XacmlContextTypeError
from ndg.xacml.core.functions import AbstractFunction, FunctionClassFactoryBase
from ndg.xacml.core.attributevalue import AttributeValueClassFactory


class MatchBase(AbstractFunction):
    """Generic match function for all types
    
    @cvar TYPE: attribute type for the given implementation.  Derived classes
    should set appropriately
    @type TYPE: NoneType    
    """
    TYPE = None
    
    def evaluate(self, input1, input2):
        """Match two inputs of type self.__class__.TYPE
        
        @param input1: first of two inputs to match
        @type input1: self.__class__.TYPE
        @param input2: second input
        @type input2: self.__class__.TYPE
        @return: True if inputs match, False otherwise
        @rtype: bool
        """
        if not isinstance(input1, self.__class__.TYPE):
            raise XacmlContextTypeError('Expecting %r derived type for '
                                        '"input1"; got %r' %
                                        (self.__class__.TYPE, 
                                         type(input1)))
            
        if not isinstance(input2, self.__class__.TYPE):
            raise XacmlContextTypeError('Expecting %r derived type for '
                                        '"input2"; got %r' %
                                        (self.__class__.TYPE, 
                                         type(input2)))
            
        return input1 == input2
 
    
attributeValueClassFactory = AttributeValueClassFactory()


class Rfc822NameMatch(AbstractFunction):
    """Match RFC 822 type name - e.g. e-mail addresses"""
    FUNCTION_NS = 'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-match'
    TYPE = attributeValueClassFactory(
                                'http://www.w3.org/2001/XMLSchema#Rfc822Name')
    STRING_TYPE = attributeValueClassFactory(
                                'http://www.w3.org/2001/XMLSchema#string')
    
    def evaluate(self, string1, rfc822Name):
        """Match Rfc822 name
        
        @param string1: string to match to rfc822Name
        @type string1: basestring
        @param rfc822Name: RFC822 Name to match
        @type rfc822Name: basestring
        @return: True if strings match, False otherwise
        @rtype: bool
        """
        if not isinstance(string1, self.__class__.STRING_TYPE):
            raise TypeError('Expecting %r derived type for "string1"; got %r' %
                            (self.__class__.STRING_TYPE, type(string1)))
            
        if not isinstance(rfc822Name, self.__class__.TYPE):
            raise TypeError('Expecting %r derived type for "rfc822Name"; got '
                            '%r' % (self.__class__.TYPE, type(rfc822Name)))
            
        return string1.lower() == rfc822Name.lower()
        
    
class FunctionClassFactory(FunctionClassFactoryBase):
    """Class Factory for *-match XACML 1.0 function classes
    
    @cvar FUNCTION_NAMES: equal function URNs
    @type FUNCTION_NAMES: tuple
    
    @cvar FUNCTION_NS_SUFFIX: generic suffix for match function URNs
    @type FUNCTION_NS_SUFFIX: string
    
    @cvar FUNCTION_BASE_CLASS: base class for all match function classes (apart
    from Rfc822NameMatch)
    @type FUNCTION_BASE_CLASS: ndg.xacml.core.functions.v1.MatchBase
    """
    FUNCTION_NAMES = (
        'urn:oasis:names:tc:xacml:1.0:function:x500Name-match',
        'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-match',
        'urn:oasis:names:tc:xacml:1.0:function:xpath-node-match'
    )
    FUNCTION_NS_SUFFIX = '-match'
    FUNCTION_BASE_CLASS = MatchBase



"""NDG XACML module for Result type 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "23/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: result.py 7443 2010-09-03 13:42:43Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core.context import XacmlContextBase
from ndg.xacml.core.obligation import Obligation


class StatusCode(XacmlContextBase):
    '''XACML Response Result StatusCode.
    
    @cvar ELEMENT_LOCAL_NAME: XML Local Name of StatusCode element 
    @type ELEMENT_LOCAL_NAME: string
    
    @cvar IDENTIFIER_PREFIX: namespace prefix for status codes
    @type IDENTIFIER_PREFIX: string
    
    @cvar OK: OK response status
    @type OK: string
    
    @cvar MISSING_ATTRIBUTE: missing attribute response status
    @type MISSING_ATTRIBUTE: string
    
    @cvar PROCESSING_ERROR: response status indicating a processing error
    @type PROCESSING_ERROR: string
    
    @cvar SYNTAX_ERROR: response status for a syntax error
    @type SYNTAX_ERROR: string
    
    @cvar CODES: list of recognised codes
    @type CODES: tuple
    
    @ivar __value: status code value
    @type __value: basestring
    
    @ivar __childStatusCode: child status code
    @type __childStatusCode: ndg.xacml.core.result.StatusCode
    '''
    
    ELEMENT_LOCAL_NAME = "StatusCode"

    IDENTIFIER_PREFIX = XacmlContextBase.XACML_1_0_NS_PREFIX + ':status'
    
    OK = IDENTIFIER_PREFIX + ":ok"
    MISSING_ATTRIBUTE = IDENTIFIER_PREFIX + ":missing-attribute"
    PROCESSING_ERROR = IDENTIFIER_PREFIX + ":processing-error"
    SYNTAX_ERROR = IDENTIFIER_PREFIX + ":syntax-error"
    
    CODES = (OK, MISSING_ATTRIBUTE, PROCESSING_ERROR, SYNTAX_ERROR)
    
    __slots__ = ('__value', '__childStatusCode',)
    
    def __init__(self, **kw):
        super(StatusCode, self).__init__(**kw)
        
        # Value attribute URI.
        self.__value = None
    
        # Nested secondary StatusCode child element.
        self.__childStatusCode = None

    def _getStatusCode(self): 
        '''@return: child status code
        @rtype: ndg.xacml.core.result.StatusCode
        '''
        return self.__childStatusCode
    
    def _setStatusCode(self, value):
        '''@param value: child status code
        @type value: ndg.xacml.core.result.StatusCode
        '''
        if not isinstance(value, StatusCode):
            raise TypeError('Child "statusCode" must be a %r derived type, '
                            "got %r" % (StatusCode, type(value)))
            
        self.__childStatusCode = value

    value = property(fget=_getStatusCode, 
                     fset=_setStatusCode, 
                     doc="Child Status code")
              
    def _getValue(self):
        '''@return: status code value
        @rtype: basestring
        '''
        return self.__value
        
    def _setValue(self, value):
        '''@param value: status code value
        @type value: basestring
        '''
        if not isinstance(value, basestring):
            raise TypeError("\"value\" must be a basestring derived type, "
                            "got %r" % value.__class__)
        
        if value not in self.__class__.CODES:
            raise AttributeError('Status code expected values are %r; got %r' %
                                 (self.__class__.CODES, value))
                
        self.__value = value

    value = property(fget=_getValue, fset=_setValue, doc="Status code value")
    

class Status(XacmlContextBase): 
    '''XACML Response Result Status

    @cvar ELEMENT_LOCAL_NAME: XML Local Name of Status element 
    @type ELEMENT_LOCAL_NAME: string
    
    @ivar __statusCode: Status code element
    @type __statusCode: None / ndg.xacml.core.context.result.StatusCode
    @ivar __statusMessage: Status message element
    @type __statusMessage: None / basestring
    @ivar __statusDetail: Status detail element
    @type __statusDetail: None / any type
    '''
    
    # Local Name of Status.
    ELEMENT_LOCAL_NAME = "Status"

    __slots__ = ('__statusCode', '__statusMessage', '__statusDetail', )
    
    def __init__(self, **kw):
        super(Status, self).__init__(**kw)
        
        # StatusCode element.
        self.__statusCode = None
    
        # StatusMessage element.
        self.__statusMessage = None
    
        # StatusDetail element. 
        self.__statusDetail = None
    
    @classmethod
    def createInitialised(cls, code=StatusCode.OK, message='', detail=''):
        """Create with an empty StatusCode object set

        @param code: Status code - defaults to OK code
        @type code: basestring
        @param message: Status Message
        @type message: basestring
        @param detail: Status detail
        @type detail: string / any
        @return: Status instance
        @rtype: ndg.xacml.core.context.Status
        """
        status = cls()
        status.statusCode = StatusCode()
        status.statusCode.value = code
        status.statusMessage = message
        status.statusDetail = detail
        
        return status
        
    def _getStatusCode(self):
        '''
        Gets the Code of this Status.
        
        @return: Status StatusCode
        @rtype: ndg.xacml.core.context.StatusCode
        '''
        return self.__statusCode

    def _setStatusCode(self, value):
        '''
        Sets the Code of this Status.
        
        @param value: the Code of this Status
        @type value: ndg.xacml.core.context.StatusCode
        '''
        if not isinstance(value, StatusCode):
            raise TypeError('"statusCode" must be a %r derived type, '
                            "got %r" % (StatusCode, type(value)))
            
        self.__statusCode = value
        
    statusCode = property(fget=_getStatusCode,
                          fset=_setStatusCode,
                          doc="status code object")
    
    def _getStatusMessage(self):
        '''
        Gets the Message of this Status.
        
        @return: Status StatusMessage
        @rtype: basestring
        '''
        return self.__statusMessage

    def _setStatusMessage(self, value):
        '''
        Sets the Message of this Status.
        
        @param value: the Message of this Status
        @type value: basestring
        '''
        if not isinstance(value, basestring):
            raise TypeError('"statusMessage" must be a %r derived type, '
                            "got %r" % (basestring, type(value)))
            
        self.__statusMessage = value
        
    statusMessage = property(fget=_getStatusMessage,
                             fset=_setStatusMessage,
                             doc="status message")

    def _getStatusDetail(self):
        '''
        Gets the Detail of this Status.
        
        @return: Status detail
        @rtype: any
        '''
        return self.__statusDetail
    
    def _setStatusDetail(self, value):
        '''
        Sets the Detail of this Status.
        
        @param value: the Detail of this Status
        @type value: any type
        '''
        self.__statusDetail = value
        
    statusDetail = property(fget=_getStatusDetail,
                            fset=_setStatusDetail,
                            doc="status message")
            

class Decision(object):
    """Define decision types for Response Result
    
    @cvar PERMIT_STR: permit decision string
    @type PERMIT_STR: string
    
    @cvar DENY_STR: deny decision string
    @type DENY_STR: string
    
    @cvar INDETERMINATE_STR: indeterminate decision string
    @type INDETERMINATE_STR: string
    
    @cvar NOT_APPLICABLE_STR: not applicable decision string
    @type NOT_APPLICABLE_STR: string
    
    @cvar TYPES: list of valid decision strings
    @type TYPES: tuple
    
    @ivar __value: decision value
    @type __value: string
    """
    
    # "Permit" decision type
    PERMIT_STR = "Permit"
    
    # "Deny" decision type
    DENY_STR = "Deny"
    
    # "Indeterminate" decision type
    INDETERMINATE_STR = "Indeterminate"
    
    NOT_APPLICABLE_STR = "NotApplicable"
        
    TYPES = (PERMIT_STR, DENY_STR, INDETERMINATE_STR, NOT_APPLICABLE_STR)
    
    __slots__ = ('__value',)
    
    def __init__(self, decision=INDETERMINATE_STR):
        self.__value = None
        self.value = decision

    def __getstate__(self):
        '''Enable pickling
        
        @return: class instance attributes dictionary
        @rtype: dict
        '''
        _dict = {}
        for attrName in Decision.__slots__:
            # Ugly hack to allow for derived classes setting private member
            # variables
            if attrName.startswith('__'):
                attrName = "_Decision" + attrName
                
            _dict[attrName] = getattr(self, attrName)
            
        return _dict
  
    def __setstate__(self, attrDict):
        '''Enable pickling
        
        @param attrDict: class instance attributes dictionary
        @type attrDict: dict
        '''
        for attrName, val in attrDict.items():
            setattr(self, attrName, val)
            
    def _setValue(self, value):
        """Set decision value
        
        @param value: decision value - constrained vocabulary to Decision.TYPES
        @type value: string or ndg.xacml.core.context.result.Decision
        @raise AttributeError: invalid decision string value input
        @raise TypeError: invalid type for input decision value
        """
        if isinstance(value, Decision):
            # Cast to string
            value = str(value)
            
        elif not isinstance(value, basestring):
            raise TypeError('Expecting string or Decision instance for '
                            '"value" attribute; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable decision types are %r; got '
                                 '%r instead' % (Decision.TYPES, value))
        self.__value = value
        
    def _getValue(self):
        """Get decision value
        
        @return: decision value 
        @rtype: string
        """
        return self.__value
    
    value = property(fget=_getValue, fset=_setValue, doc="Decision value")
    
    def __str__(self):
        """represent decision as a string
        
        @return: decision value 
        @rtype: string 
        """
        return self.__value

    def __repr__(self):
        """Overridden to show the decision value
        
        @return: decision representation 
        @rtype: string 
        """
        return "%s = %r" % (super(Decision, self).__repr__(), self.__value)
    
    def __eq__(self, decision):
        """
        @param decision: decision value to compare with self's
        @type decision: string or ndg.xacml.core.context.result.Decision
        @return: True if the decision values match, False otherwise
        @rtype: bool
        @raise AttributeError: invalid decision string value input
        @raise TypeError: invalid type for input decision value
        """
        if isinstance(decision, Decision):
            # Cast to string
            value = decision.value
            
        elif isinstance(decision, basestring):
            value = decision
            
        else:
            raise TypeError('Expecting string or Decision instance for '
                            'input decision value; got %r instead' % type(value))
            
        if value not in self.__class__.TYPES:
            raise AttributeError('Permissable decision types are %r; got '
                                 '%r instead' % (Decision.TYPES, value))
            
        return self.__value == value   


class PermitDecision(Decision):
    """Permit authorisation Decision"""
    __slots__ = ()

    def __init__(self):
        """Initialise set with Permit value"""
        super(PermitDecision, self).__init__(Decision.PERMIT_STR)
        
    def _setValue(self):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")


class DenyDecision(Decision):
    """Deny authorisation Decision"""
    __slots__ = ()
    
    def __init__(self):
        """Initialise set with deny value"""
        super(DenyDecision, self).__init__(Decision.DENY_STR)
        
    def _setValue(self, value):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")


class IndeterminateDecision(Decision):
    """Indeterminate authorisation Decision"""
    __slots__ = ()
    
    def __init__(self):
        """Initialise set with indeterminate value"""
        super(IndeterminateDecision, self).__init__(Decision.INDETERMINATE_STR)
        
    def _setValue(self, value):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")


class NotApplicableDecision(Decision):
    """NotApplicable authorisation Decision"""
    __slots__ = ()
    
    def __init__(self):
        """Initialise set with not applicable value"""
        super(NotApplicableDecision, self).__init__(Decision.NOT_APPLICABLE_STR)
        
    def _setValue(self, value):  
        """Make value read-only
        @raise AttributeError: value can't be set
        """
        raise AttributeError("can't set attribute")
    
    
# Add instances of each for convenience
Decision.PERMIT = PermitDecision()
Decision.DENY = DenyDecision()
Decision.INDETERMINATE = IndeterminateDecision()
Decision.NOT_APPLICABLE = NotApplicableDecision()


class Result(XacmlContextBase):
    """XACML Result type - element in a Response
    
    @cvar ELEMENT_LOCAL_NAME: XML element local name
    @type ELEMENT_LOCAL_NAME: string
    
    @cvar OBLIGATIONS_ELEMENT_LOCAL_NAME: obligations XML element local name
    @type OBLIGATIONS_ELEMENT_LOCAL_NAME: string
    
    @cvar RESOURCE_ID_ATTRIB_NAME: resource ID XML attribute name
    @type RESOURCE_ID_ATTRIB_NAME: string
    
    @ivar __resourceId: resource id
    @type __resourceId: None/basestring
    
    @ivar __decision: decision for this result
    @type __decision: ndg.xacml.core.context.result.Decision
    
    @ivar __status: result status
    @type __status: ndg.xacml.core.context.result.Status
       
    @ivar __obligations: obligations associated with this result
    @type __obligations: ndg.xacml.core.obligation.Obligation
    """
    __slots__ = ('__resourceId', '__decision', '__status', '__obligations')
    
    ELEMENT_LOCAL_NAME  = 'Result'
    OBLIGATIONS_ELEMENT_LOCAL_NAME = 'Obligations'
    RESOURCE_ID_ATTRIB_NAME = 'ResourceId'
    
    def __init__(self):
        super(Result, self).__init__()
        self.__decision = None
        self.__status = None
        self.__resourceId = None
        self.__obligations = None
        
    @classmethod
    def createInitialised(cls, 
                          decision=Decision.NOT_APPLICABLE,
                          resourceId='', 
                          obligations=None,
                          **kw):
        """Create a result object populated with all it's child elements
        rather than set to None as is the default from __init__
        
        @param decision: decision for this result
        @type decision: ndg.xacml.core.context.result.Decision
        
        @param resourceId: resource id for associated resource
        @type resourceId: basestring
        
        @param obligations: obligations associated with this result
        @type obligations: None/ndg.xacml.core.obligation.Obligation
        
        @param kw: keywords for status attribute initialisation
        @type kw: dict
        
        @return: new result object with all its child attributes created
        @rtype: ndg.xacml.core.context.result.Result
        """
        result = cls()
        result.decision = Decision()
        result.decision.value = decision
        result.status = Status.createInitialised(**kw)
        
        if obligations is not None:
            result.obligations = obligations
            
        return result
        
    @property
    def resourceId(self):
        """Get Result resource Id
        
        @return: resource Id
        @rtype: basestring
        """
        return self.__resourceId

    @resourceId.setter
    def resourceId(self, value):
        """Set Result resource Id
        
        @param value: resource Id
        @type value: basestring
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "resourceId" '
                            'result; got %r' % (basestring, type(value)))
            
        self.__resourceId = value
                        
    @property
    def decision(self):
        """Get Result decision
        
        @return: decision for this result
        @rtype: ndg.xacml.core.context.result.Decision
        """
        return self.__decision
    
    @decision.setter
    def decision(self, value):
        """Set Request decision
        
        @param value: decision for this result
        @type value: ndg.xacml.core.context.result.Decision
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, Decision):
            raise TypeError('Expecting %r type for result "decision" '
                            'attribute; got %r' % (Decision, type(value)))
        self.__decision = value
        
    @property
    def status(self):
        """Get Result status
        
        @return: status for this result
        @rtype: ndg.xacml.core.context.result.Status
        """
        return self.__status
    
    @status.setter
    def status(self, value):
        """Set Result status
        
        @param value: status for this result
        @type value: ndg.xacml.core.context.result.Status
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, Status):
            raise TypeError('Expecting %r type for result "status" '
                            'attribute; got %r' % (Status, type(value)))
            
        self.__status = value
                                    
    @property
    def obligations(self):
        """Get Result obligations
        
        @return: obligations associated with this result
        @rtype: ndg.xacml.core.obligations.Obligations        
        """
        return self.__obligations
    
    @obligations.setter
    def obligations(self, value):
        """Set Result obligations
        
        @param value: obligations associated with this result
        @type value: ndg.xacml.core.obligations.Obligations        
        
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, Obligation):
            raise TypeError('Expecting %r type for result "obligations" '
                            'attribute; got %r' % (Obligation, type(value)))
            
        self.__obligations = value


"""NDG XACML Context Subject type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: subject.py 7087 2010-06-25 11:23:09Z pjkersha $"
from ndg.xacml.core.context import RequestChildBase


class Subject(RequestChildBase):
    """XACML Context Subject type
    
    @cvar ELEMENT_LOCAL_NAME: XML Local Name of this element 
    @type ELEMENT_LOCAL_NAME: string
    
    @cvar SUBJECT_CATEGORY_ATTRIB_NAME: subject category XML attribute name
    @type SUBJECT_CATEGORY_ATTRIB_NAME: string
    
    @ivar __subjectCategory: subject category XML attribute name
    @type __subjectCategory: string    
    """
    ELEMENT_LOCAL_NAME = 'Subject'
    SUBJECT_CATEGORY_ATTRIB_NAME = 'SubjectCategory'
    
    __slots__ = ('__subjectCategory',)
    
    def __init__(self):
        super(Subject, self).__init__()
        self.__subjectCategory = None

    def _get_subjectCategory(self):
        """Get subject category
        
        @return: subject category XML attribute name
        @rtype: string 
        """   
        return self.__subjectCategory

    def _set_subjectCategory(self, value):
        """Set subject category
        
        @param value: subject category XML attribute name
        @type value: string    
        
        @raise TypeError: incorrect type for input
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "subjectCategory" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__subjectCategory = value   

    subjectCategory = property(_get_subjectCategory, _set_subjectCategory, None, 
                               "Subject category")


"""NDG XACML package for functions

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "26/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7351 2010-08-20 14:46:35Z pjkersha $"
from abc import ABCMeta, abstractmethod
from datetime import datetime, timedelta
import traceback
import logging
log = logging.getLogger(__name__)

from ndg.xacml.core.attributevalue import (AttributeValue, 
                                           AttributeValueClassFactory)
from ndg.xacml.utils import VettedDict, _isIterable
from ndg.xacml.utils.factory import callModuleObject


class AbstractFunction(object):
    """Abstract Base class for all XACML matching functions
    @cvar FUNCTION_NS: namespace for the given function
    @type FUNCTION_NS: NoneType (must be string in derived type)
    
    @cvar V1_0_FUNCTION_NS: XACML 1.0 function namespace prefix
    @type V1_0_FUNCTION_NS: string
    
    @cvar V2_0_FUNCTION_NS: XACML 2.0 function namespace prefix
    @type V2_0_FUNCTION_NS: string
    """
    __metaclass__ = ABCMeta
    
    FUNCTION_NS = None
    V1_0_FUNCTION_NS = "urn:oasis:names:tc:xacml:1.0:function:"
    V2_0_FUNCTION_NS = "urn:oasis:names:tc:xacml:2.0:function:"
    
    def __init__(self):
        """
        @raise TypeError: if FUNCTION_NS not set correctly
        """
        if self.__class__.FUNCTION_NS is None:
            raise TypeError('"FUNCTION_NS" class variable must be defined in '
                            'derived classes')
            
    @abstractmethod
    def evaluate(self, *inputs):
        """Evaluate the function from the given input arguments and context
        
        @param inputs: input arguments need to evaluate the function
        @type inputs: tuple
        @return: derived type should return True for match, False otherwise
        @rtype: bool (derived type), NoneType for THIS implementation
        """
        
class XacmlFunctionNames(object):
    """XACML standard match function names
    
    @cvar FUNCTION_NAMES: list of all the XACML function URNs
    @type FUNCTION_NAMES: tuple    
    """
    FUNCTION_NAMES = (
'urn:oasis:names:tc:xacml:1.0:function:string-equal',
'urn:oasis:names:tc:xacml:1.0:function:boolean-equal',
'urn:oasis:names:tc:xacml:1.0:function:integer-equal',
'urn:oasis:names:tc:xacml:1.0:function:double-equal',
'urn:oasis:names:tc:xacml:1.0:function:date-equal',
'urn:oasis:names:tc:xacml:1.0:function:time-equal',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-equal',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-equal',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-equal',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-equal',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-equal',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-equal',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-equal',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-equal',
'urn:oasis:names:tc:xacml:1.0:function:integer-add',
'urn:oasis:names:tc:xacml:1.0:function:double-add',
'urn:oasis:names:tc:xacml:1.0:function:integer-subtract',
'urn:oasis:names:tc:xacml:1.0:function:double-subtract',
'urn:oasis:names:tc:xacml:1.0:function:integer-multiply',
'urn:oasis:names:tc:xacml:1.0:function:double-multiply',
'urn:oasis:names:tc:xacml:1.0:function:integer-divide',
'urn:oasis:names:tc:xacml:1.0:function:double-divide',
'urn:oasis:names:tc:xacml:1.0:function:integer-mod',
'urn:oasis:names:tc:xacml:1.0:function:integer-abs',
'urn:oasis:names:tc:xacml:1.0:function:double-abs',
'urn:oasis:names:tc:xacml:1.0:function:round',
'urn:oasis:names:tc:xacml:1.0:function:floor',
'urn:oasis:names:tc:xacml:1.0:function:string-normalize-space',
'urn:oasis:names:tc:xacml:1.0:function:string-normalize-to-lower-case',
'urn:oasis:names:tc:xacml:1.0:function:double-to-integer',
'urn:oasis:names:tc:xacml:1.0:function:integer-to-double',
'urn:oasis:names:tc:xacml:1.0:function:or',
'urn:oasis:names:tc:xacml:1.0:function:and',
'urn:oasis:names:tc:xacml:1.0:function:n-of',
'urn:oasis:names:tc:xacml:1.0:function:not',
'urn:oasis:names:tc:xacml:1.0:function:integer-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:integer-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:integer-less-than',
'urn:oasis:names:tc:xacml:1.0:function:integer-less-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:double-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:double-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:double-less-than',
'urn:oasis:names:tc:xacml:1.0:function:double-less-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-add-dayTimeDuration',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-add-yearMonthDuration',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-subtract-dayTimeDuration',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-subtract-yearMonthDuration', 
'urn:oasis:names:tc:xacml:1.0:function:date-add-yearMonthDuration',
'urn:oasis:names:tc:xacml:1.0:function:date-subtract-yearMonthDuration',
'urn:oasis:names:tc:xacml:1.0:function:string-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:string-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:string-less-than',
'urn:oasis:names:tc:xacml:1.0:function:string-less-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:time-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:time-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:time-less-than',
'urn:oasis:names:tc:xacml:1.0:function:time-less-than-or-equal',
'urn:oasis:names:tc:xacml:2.0:function:time-in-range',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-less-than',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-less-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:date-greater-than',
'urn:oasis:names:tc:xacml:1.0:function:date-greater-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:date-less-than',
'urn:oasis:names:tc:xacml:1.0:function:date-less-than-or-equal',
'urn:oasis:names:tc:xacml:1.0:function:string-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:string-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:string-is-in',
'urn:oasis:names:tc:xacml:1.0:function:string-bag',
'urn:oasis:names:tc:xacml:1.0:function:boolean-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:boolean-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:boolean-is-in',
'urn:oasis:names:tc:xacml:1.0:function:boolean-bag',
'urn:oasis:names:tc:xacml:1.0:function:integer-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:integer-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:integer-is-in',
'urn:oasis:names:tc:xacml:1.0:function:integer-bag',
'urn:oasis:names:tc:xacml:1.0:function:double-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:double-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:double-is-in',
'urn:oasis:names:tc:xacml:1.0:function:double-bag',
'urn:oasis:names:tc:xacml:1.0:function:time-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:time-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:time-is-in',
'urn:oasis:names:tc:xacml:1.0:function:time-bag',
'urn:oasis:names:tc:xacml:1.0:function:date-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:date-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:date-is-in',
'urn:oasis:names:tc:xacml:1.0:function:date-bag',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-is-in',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-bag',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-is-in',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-bag',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-is-in',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-bag',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-is-in',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-bag',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-is-in',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-bag',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-is-in',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-bag',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-is-in',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-bag',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-one-and-only',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-bag-size',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-is-in',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-bag',
'urn:oasis:names:tc:xacml:2.0:function:string-concatenate',
'urn:oasis:names:tc:xacml:2.0:function:uri-string-concatenate',
'urn:oasis:names:tc:xacml:1.0:function:any-of',
'urn:oasis:names:tc:xacml:1.0:function:all-of',
'urn:oasis:names:tc:xacml:1.0:function:any-of-any',
'urn:oasis:names:tc:xacml:1.0:function:all-of-any',
'urn:oasis:names:tc:xacml:1.0:function:any-of-all',
'urn:oasis:names:tc:xacml:1.0:function:all-of-all',
'urn:oasis:names:tc:xacml:1.0:function:map',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-match',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-match',
'urn:oasis:names:tc:xacml:1.0:function:string-regexp-match',
'urn:oasis:names:tc:xacml:2.0:function:anyURI-regexp-match',
'urn:oasis:names:tc:xacml:2.0:function:ipAddress-regexp-match',
'urn:oasis:names:tc:xacml:2.0:function:dnsName-regexp-match',
'urn:oasis:names:tc:xacml:2.0:function:rfc822Name-regexp-match',
'urn:oasis:names:tc:xacml:2.0:function:x500Name-regexp-match',
'urn:oasis:names:tc:xacml:1.0:function:xpath-node-count',
'urn:oasis:names:tc:xacml:1.0:function:xpath-node-equal',
'urn:oasis:names:tc:xacml:1.0:function:xpath-node-match',
'urn:oasis:names:tc:xacml:1.0:function:string-intersection',
'urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:string-union',
'urn:oasis:names:tc:xacml:1.0:function:string-subset',
'urn:oasis:names:tc:xacml:1.0:function:string-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:boolean-intersection',
'urn:oasis:names:tc:xacml:1.0:function:boolean-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:boolean-union',
'urn:oasis:names:tc:xacml:1.0:function:boolean-subset',
'urn:oasis:names:tc:xacml:1.0:function:boolean-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:integer-intersection',
'urn:oasis:names:tc:xacml:1.0:function:integer-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:integer-union',
'urn:oasis:names:tc:xacml:1.0:function:integer-subset',
'urn:oasis:names:tc:xacml:1.0:function:integer-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:double-intersection',
'urn:oasis:names:tc:xacml:1.0:function:double-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:double-union',
'urn:oasis:names:tc:xacml:1.0:function:double-subset',
'urn:oasis:names:tc:xacml:1.0:function:double-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:time-intersection',
'urn:oasis:names:tc:xacml:1.0:function:time-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:time-union',
'urn:oasis:names:tc:xacml:1.0:function:time-subset',
'urn:oasis:names:tc:xacml:1.0:function:time-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:date-intersection',
'urn:oasis:names:tc:xacml:1.0:function:date-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:date-union',
'urn:oasis:names:tc:xacml:1.0:function:date-subset',
'urn:oasis:names:tc:xacml:1.0:function:date-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-intersection',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-union',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-subset',
'urn:oasis:names:tc:xacml:1.0:function:dateTime-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-intersection',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-union',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-subset',
'urn:oasis:names:tc:xacml:1.0:function:anyURI-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-intersection',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-union',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-subset',
'urn:oasis:names:tc:xacml:1.0:function:hexBinary-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-intersection',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-union',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-subset',
'urn:oasis:names:tc:xacml:1.0:function:base64Binary-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-intersection',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-union',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-subset',
'urn:oasis:names:tc:xacml:1.0:function:dayTimeDuration-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-intersection',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-union',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-subset',
'urn:oasis:names:tc:xacml:1.0:function:yearMonthDuration-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-intersection',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-union',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-subset',
'urn:oasis:names:tc:xacml:1.0:function:x500Name-set-equals',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-intersection',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-at-least-one-member-of',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-union',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-subset',
'urn:oasis:names:tc:xacml:1.0:function:rfc822Name-set-equals',
    )

from ndg.xacml import XacmlError


class UnsupportedFunctionError(XacmlError):
    """Encountered a function type that is not recognised as part of the XACML
    specification and is not supported in this implementation"""

 
class UnsupportedStdFunctionError(UnsupportedFunctionError): 
    """Encountered a function type that is not supported even though it is
    part of the XACML specification"""
    
    
def unsupportedFunctionErrorFactory(identifier, msg=None):
    """Factory function to return an unsupported function exception based on
    the function identifier passed in
    
    @param identifier: XACML function namespace to check
    @type identifier: basestring
    
    @return: unsupported function exception instance
    @rtype: UnsupportedFunctionError or UnsupportedStdFunctionError depending
    on the identifier passed
    """
    if identifier in XacmlFunctionNames.FUNCTION_NAMES:
        if msg is None:
            msg = "%s: %s" % (UnsupportedStdFunctionError.__doc__, identifier)
            
        raise UnsupportedStdFunctionError(msg)
    else:
        if msg is None:
            msg = "%s: %s" % (UnsupportedFunctionError.__doc__, identifier)
            
        raise UnsupportedFunctionError(msg)
    
    
class FunctionClassFactoryInterface(object):
    """Interface class for function module class factory class
    """
    __meta__ = ABCMeta
    
    @abstractmethod
    def __call__(self, identifier):
        '''Create class for the given XACML function identifier
        
        @param identifier: XACML function identifier
        @type identifier: basestring
        @return: at least one member of class corresponding to the given input
        identifier
        @rtype: AbstractFunction derived type or NoneType if no match is found
        '''
        return None
    

class FunctionClassFactoryBase(FunctionClassFactoryInterface):
    """Base implementation for XACML Function Class Factory.  There should be
    one derived type for each function family implemented in sub-modules of 
    ndg.xacml.core.functions 
    
    e.g.
    
    for urn:oasis:names:tc:xacml:1.0:function:<type>-at-least-one-member-of a 
    class factory should exist,
    
    ndg.xacml.core.functions.v1.at_least_one_member_of.FunctionClassFactory
    
    which will be capable of returning a type derived from AbstractFunction:
    
    <type>AtLeastOneMemberOf    
    
    e.g. StringAtLeastOneMemberOf, BooleanAtLeastOneMemberOf.
    
    This class is for convenience only some function factories are better 
    derived directly from FunctionClassFactoryInterface
    
    Derived classes MUST define these class variables:
    
    @cvar FUNCTION_NAMES: list of function identifiers that this factory can
    produce classes for e.g.:
    
    ('urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of', ...)
    
    @type FUNCTION_NAMES: NoneType (but list in derived class)
   
    @cvar FUNCTION_NS_SUFFIX: urn suffix for the family of function to define
    e.g. -at-least-one-member-of is the suffix for the URN:
    
    urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of
    @type FUNCTION_NS_SUFFIX: NoneType (but basestring in derived class)
    
    @cvar FUNCTION_BASE_CLASS: base class for this family of functions e.g for
    urn:oasis:names:tc:xacml:1.0:function:string-at-least-one-member-of,
    ndg.xacml.core.functions.v1.at_least_one_member_of.AtLeastOneMemberOfBase
    @type FUNCTION_BASE_CLASS: NoneType (but AbstractFunction derived type in 
    derived function factory class)
    """
    
    FUNCTION_NS_SUFFIX = None
    FUNCTION_NAMES = None
    FUNCTION_BASE_CLASS = None
    
    URN_SEP = ':'
    FUNCTION_NAME_SEP = '-'
    __slots__ = ('__map', 'attributeValueClassFactory', 'functionSuffix')
    
    def __init__(self):
        '''This class is in fact abstract - derived types must define the 
        FUNCTION_NS_SUFFIX and FUNCTION_BASE_CLASS class variables
        '''
        if None in (self.__class__.FUNCTION_NS_SUFFIX, 
                    self.__class__.FUNCTION_BASE_CLASS):
            raise TypeError('"FUNCTION_NS_SUFFIX" and "FUNCTION_BASE_CLASS" '
                            'must be defined in a derived implementation of '
                            'FunctionClassFactoryBase.  See '
                            'FunctionClassFactoryBase.__doc__ contents')
        
        if not _isIterable(self.__class__.FUNCTION_NAMES):
            raise TypeError('"FUNCTION_NAMES" class variable must be an '
                            'iterable of string type function identifiers; got '
                            '%r' % self.__class__.FUNCTION_NAMES)

        self.__map = {}   
        
        # Enables creation of matching attribute types to relevant to the 
        # function classes    
        self.attributeValueClassFactory = AttributeValueClassFactory()
            
        
        functionSuffixParts = self.__class__.FUNCTION_NS_SUFFIX.split(
                                            self.__class__.FUNCTION_NAME_SEP)
        self.functionSuffix = ''.join([n[0].upper() + n[1:] 
                                  for n in functionSuffixParts if n])
        
    def initAllFunctionClasses(self):
        """Create classes for all functions for a data type e.g. a derived class
        could implement a factory for <type>-at-least-one-member-of functions:
        string-at-least-one-member-of, boolean-at-least-one-member-of, etc. 
        
        Function classes are placed in a look-up table __map for the __call__()
        method to access
        
        In practice, there shouldn't be a need to load all the functions in
        one go.  The __call__ method loads functions and caches them as needed.
        """        
        for identifier in self.__class__.FUNCTION_NAMES:
            self.loadFunction(identifier)        

    def loadFunction(self, identifier):
        """Create a class for the given function namespace and cache it in the 
        function class look-up table for future requests.  Note that this call
        overwrites any existing entry in the cache whereas __call__ will try
        to use an entry in the cache if it already exists
        
        @param identifier: XACML function namespace
        @type identifier: basestring
        """

        # str.capitalize doesn't do what's required: need to capitalize the 
        # first letter of the word BUT retain camel case for the rest of it
        _capitalize = lambda s: s[0].upper() + s[1:]
        
        # Extract the function name and the type portion of the function
        # name in order to make an implementation of a class to handle it
        functionName = identifier.split(self.__class__.URN_SEP)[-1]
        typePart = functionName.split(self.__class__.FUNCTION_NS_SUFFIX)[0]
        
        # Attempt to infer from the function name the associated type
        typeName = _capitalize(typePart)
        
        # Remove any hyphens converting to camel case
        if '-' in typeName:
            typeName = ''.join([_capitalize(i) for i in typeName.split('-')])
            
        typeURI = AttributeValue.TYPE_URI_MAP.get(typeName)
        if typeURI is None:
            # Ugly hack to allow for XPath node functions
            if typePart == 'xpath-node':
                typeURI = AttributeValue.TYPE_URI_MAP['String']
            else:
                raise TypeError('No AttributeValue.TYPE_URI_MAP entry for '
                                '%r type' % typePart) 
            
        _type = self.attributeValueClassFactory(typeURI)
        if _type is None:
            raise TypeError('No AttributeValue.TYPE_MAP entry for %r type' %
                            typeName)
          
        className = typeName + self.functionSuffix
        classVars = {
            'TYPE': _type,
            'FUNCTION_NS': identifier
        }
        
        functionClass = type(className, 
                             (self.__class__.FUNCTION_BASE_CLASS, ), 
                             classVars)
        
        self.__map[identifier] = functionClass
            
    def __call__(self, identifier):
        """Return the class for the given XACML type function identifier
        
        @param identifier: XACML *-at-least-one-member-of type function
        identifier
        @type identifier: basestring
        @return: at least one member of class corresponding to the given input
        identifier
        @rtype: AtLeastOneMemberOfBase derived type or None if no match is 
        found
        """
        # Check the cache first
        functionClass = self.__map.get(identifier)
        if functionClass is None:
            # No class set in the cache - try loading the new class and updating
            # the cache.
            self.loadFunction(identifier)
            
        # This should result in a safe retrieval from the cache because of the
        # above check - None return would result otherwise.
        return self.__map.get(identifier)
        
    
class FunctionMapError(Exception):
    """Generic Error exception class for FunctionMap"""
    
    
class FunctionMapConfigError(FunctionMapError):
    """Configuration related exception for FunctionMap"""
        
        
class FunctionMap(VettedDict):
    """Map function IDs to their class implementations in the various function
    sub-modules.  It provide a layer over the various 
    FunctionClassFactoryInterface implementations so that a function class can 
    be obtained directly from a given XACML function URN.  
    
    @cvar FUNCTION_PKG_PREFIX: python package path for functions package
    @type FUNCTION_PKG_PREFIX: string
    
    @cvar V1_0_PKG_PREFIX: python package path for XACML 1.0 functions package
    @type V1_0_PKG_PREFIX: string
    
    @cvar V2_0_PKG_PREFIX: python package path for XACML 2.0 functions package
    @type V2_0_PKG_PREFIX: string
    
    @cvar SUPPORTED_NSS: mapping of function URN prefix to Python package
    @type SUPPORTED_NSS: dict
    
    @cvar FUNCTION_CLASS_FACTORY_CLASSNAME: standard name for class factory
    which should be present in each generic function module.  This factory is 
    invoked to create the function class for any given function URN related to
    that module
    @type FUNCTION_CLASS_FACTORY_CLASSNAME: string
    """
    FUNCTION_PKG_PREFIX = 'ndg.xacml.core.functions.'
    
    V1_0_PKG_PREFIX = FUNCTION_PKG_PREFIX + 'v1.'
    V2_0_PKG_PREFIX = FUNCTION_PKG_PREFIX + 'v2.'
    
    SUPPORTED_NSS = {
        AbstractFunction.V1_0_FUNCTION_NS: V1_0_PKG_PREFIX,
        AbstractFunction.V2_0_FUNCTION_NS: V2_0_PKG_PREFIX
    }
    
    # Each function module is expected to have a class factory for obtaining
    # a class for the given function identifier associated with that module
    FUNCTION_CLASS_FACTORY_CLASSNAME = 'FunctionClassFactory'
    
    def __init__(self):
        """Force type for dictionary key value pairs: function values must be
        of AbstractFunction derived type and ID keys string type
        """        
        # Filters are defined as staticmethods but reference via self here to 
        # enable derived class to override them as standard methods without
        # needing to redefine this __init__ method            
        VettedDict.__init__(self, self.keyFilter, self.valueFilter)
        
        # This classes maintains a list of XACML function URN -> Function class
        # mappings.  This additional dict enables caching of class factories 
        # used to obtain the function classes.  There is one class factory per
        # function module e.g. ndg.xacml.core.functions.v1.equal contains a 
        # class factory which creates the various 
        # urn:oasis:names:tc:xacml:1.0:function:<type>-equal function classes
        self.__classFactoryMap = {}
        
    @staticmethod
    def keyFilter(key):
        """Enforce string type keys
        
        @param key: function URN
        @type key: basestring
        @return: True for valid key type
        @rtype: bool
        @raise TypeError: invalid key type
        """
        if not isinstance(key, basestring):
            raise TypeError('Expecting %r type for key; got %r' % 
                            (basestring, type(key))) 
            
        return True 
    
    @staticmethod
    def valueFilter(value):
        """Enforce AbstractFunction derived types for match functions
        
        @param value: function URN
        @type value: ndg.xacml.core.functions.AbstractFunction / NotImplemented
        @return: True for valid function type
        @rtype: bool
        @raise TypeError: invlaid key type
        """
        if value is NotImplemented:
            return True
        
        elif not issubclass(value, AbstractFunction):
            raise TypeError('Expecting %r derived type for value; got %r' % 
                            (AbstractFunction, value)) 
            
        return True 
           
    def loadAll(self):
        """Load function map with implementations from the relevant function
        package"""
        
        for functionNs in XacmlFunctionNames.FUNCTION_NAMES:
            self.loadFunction(functionNs)
            
    def loadFunction(self, functionNs):
        """Get package to retrieve function class for the given XACML function
        namespace
        
        @param functionNs: XACML function namespace
        @type functionNs: basestring
        """
        functionFactory = self.__classFactoryMap.get(functionNs)
        if functionFactory is not None:
            # Get function class from previously cached factory
            self[functionNs] = functionFactory(functionNs)
            return
            
        # No Factory has been cached for this function yet
        cls = FunctionMap
        classPath = None
        
        for namespacePrefix, pkgNamePrefix in cls.SUPPORTED_NSS.items():
            if functionNs.startswith(namespacePrefix):
                # Namespace is recognised - translate into a path to a 
                # function class in the right functions package
                functionName = functionNs.split(namespacePrefix)[-1]
                functionNameParts = functionName.split('-')
                
                if len(functionNameParts) == 1:
                    moduleName = functionNameParts[0]
                    
                elif functionName.startswith('xpath-node'):
                    # Ugly hack for xpath-node functions
                    moduleName = functionNameParts[-1].lower()
                else:
                    moduleName = '_'.join(functionNameParts[1:]).lower()
                    
                classPath = pkgNamePrefix + moduleName + '.' + \
                            cls.FUNCTION_CLASS_FACTORY_CLASSNAME
                break

        if classPath is None:
            raise FunctionMapConfigError('Namespace for function not '
                                         'recognised: %r' % functionNs) 
                       
        # Try instantiating the function class and loading it into the map
        try:
            functionFactory = callModuleObject(classPath)
                      
        except (ImportError, AttributeError), e:
            log.error("Error importing function factory class %r for function "
                      "identifier %r: %s", classPath, functionNs, str(e))
            
            # No implementation exists - default to Abstract function
            self[functionNs] = NotImplemented
        else:
            function = functionFactory(functionNs)
            if function is None:
                raise unsupportedFunctionErrorFactory(functionNs)
                
            self[functionNs] = function
            self.__classFactoryMap[functionNs] = functionFactory
                       
    def __getitem__(self, key):
        """Override base class implementation to load and cache function classes
        if they don't otherwise exist
        
        @param key: function URN
        @type key: basestring
        @return: function class
        @rtype: ndg.xacml.core.functions.AbstractFunction / NotImplemented
        """
        functionClass = VettedDict.get(self, key)
        if functionClass is None:
            self.loadFunction(key)
            
        return VettedDict.__getitem__(self, key)
        
    def get(self, key, *arg):
        """Likewise to __getitem__, enable loading and caching of function 
        classes if they don't otherwise exist
        
        @param key: XACML function URN
        @type key: basestring
        @param arg: set a single additional argument if required which is 
        used as the default value should the key not be found in the map
        @type arg: tuple
        @return: function class
        @rtype: ndg.xacml.core.functions.AbstractFunction / NotImplemented
        """
        functionClass = VettedDict.get(self, key, *arg)
        if functionClass is None:
            self.loadFunction(key)
            return VettedDict.get(self, key, *arg)    
        else:
            return functionClass


        


"""NDG XACML ElementTree based generic reader for subject, resource, action and
environment match types

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: matchreader.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.core.attributevalue import AttributeValue
from ndg.xacml.core.attributeselector import AttributeSelector
from ndg.xacml.parsers import XMLParseError
from ndg.xacml.parsers.etree import QName
from ndg.xacml.parsers.etree.reader import ETreeAbstractReader
from ndg.xacml.parsers.etree.factory import ReaderFactory


class MatchReaderBase(ETreeAbstractReader):
    '''ElementTree based XACML generic Match parser for subject, resource, 
    action and environment match types
    
    @cvar ATTRIBUTE_DESIGNATOR_TYPE: type for attribute designator sub-elements:
    derived class should set to relevant type e.g. for SubjectMatch,
    SubjectAttributeDesignator
    @type ATTRIBUTE_DESIGNATOR_TYPE: NoneType
    '''
    ATTRIBUTE_DESIGNATOR_TYPE = None
    
    def __init__(self):
        """Virtual method
        
        @raise NotImplementedError: ATTRIBUTE_DESIGNATOR_TYPE must be set in a 
        derived class
        """
        if self.__class__.ATTRIBUTE_DESIGNATOR_TYPE is None:
            raise NotImplementedError('Extend this class setting the '
                                      '"ATTRIBUTE_DESIGNATOR_TYPE" class '
                                      'variable')
            
        super(MatchReaderBase, self).__init__()
        
    def __call__(self, obj):
        """Parse *Match object (where * = Subject, Resource, Environment or
        Action
        
        @param obj: input object to parse
        @type obj: ElementTree Element, or stream object
        @return: new XACML match instance
        @rtype: ndg.xacml.core.matchreader.MatchReaderBase derived type 
        @raise XMLParseError: error reading element                 
        """
        elem = super(MatchReaderBase, self)._parse(obj)
        
        xacmlType = self.__class__.TYPE
        match = xacmlType()
        
        localName = QName.getLocalPart(elem.tag)
        if localName != xacmlType.ELEMENT_LOCAL_NAME:
            raise XMLParseError("No \"%s\" element found" % 
                                xacmlType.ELEMENT_LOCAL_NAME)
        
        # Unpack *required* attributes from top-level element
        attributeValues = []
        for attributeName in (xacmlType.MATCH_ID_ATTRIB_NAME, ):
            attributeValue = elem.attrib.get(attributeName)
            if attributeValue is None:
                raise XMLParseError('No "%s" attribute found in "%s" '
                                    'element' % (attributeName,
                                                 xacmlType.ELEMENT_LOCAL_NAME))
                
            attributeValues.append(attributeValue) 
                   
        match.matchId, = attributeValues
         
        # Assign specific attribute designator type from derived class
        attributeDesignatorType = self.__class__.ATTRIBUTE_DESIGNATOR_TYPE
        attributeDesignatorReaderType = ReaderFactory.getReader(
                                                        attributeDesignatorType)
     
        # Parse match elements
        for childElem in elem:
            localName = QName.getLocalPart(childElem.tag)
            
            if localName == xacmlType.ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME:
                AttributeValueReader = ReaderFactory.getReader(AttributeValue)
                match.attributeValue = AttributeValueReader.parse(childElem)
                
            elif localName == attributeDesignatorType.ELEMENT_LOCAL_NAME:
                if match.attributeSelector is not None:
                    raise XMLParseError("XACML %r child element may only be "
                                        "either a %r or %r element NOT both" %
                            (xacmlType.ELEMENT_LOCAL_NAME,
                             attributeDesignatorType.ELEMENT_LOCAL_NAME,
                             AttributeSelector.ELEMENT_LOCAL_NAME))
                    
                match.attributeDesignator = attributeDesignatorReaderType.parse(
                                                                    childElem)
                        
            elif localName == AttributeSelector.ELEMENT_LOCAL_NAME:
                if match.attributeDesignator is not None:
                    raise XMLParseError("XACML %r child element may only be "
                                        "either a %r or %r element NOT both" %
                            (xacmlType.ELEMENT_LOCAL_NAME,
                             attributeDesignatorType.ELEMENT_LOCAL_NAME,
                             AttributeSelector.ELEMENT_LOCAL_NAME))
                    
                AttributeSelectorReader = ReaderFactory.getReader(
                                                            AttributeSelector)
                
                match.attributeSelector = AttributeSelectorReader.parse(
                                                                    childElem)
            else:
                raise XMLParseError("XACML %r child element name %r not "
                                    "recognised" % 
                                    (xacmlType.MATCH_TYPE.ELEMENT_LOCAL_NAME,
                                     localName))
                    
        return match

    
        




"""NDG Security Expression type definition

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "25/02/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: expression.py 7103 2010-06-25 21:03:35Z pjkersha $"
from abc import ABCMeta, abstractmethod

from ndg.xacml.core import XacmlCoreBase


class Expression(XacmlCoreBase):
    """XACML Expression type
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar DATA_TYPE_ATTRIB_NAME: XML attribute name for data type
    @type DATA_TYPE_ATTRIB_NAME: string
    
    @ivar __dataType: data type for this expression
    @type __dataType: None / basestring
    """
    __metaclass__ = ABCMeta
    ELEMENT_LOCAL_NAME = None
    DATA_TYPE_ATTRIB_NAME = 'DataType'
    
    __slots__ = ('__dataType', )
    
    def __init__(self):
        super(Expression, self).__init__()
        self.__dataType = None
        
    def _get_dataType(self):
        return self.__dataType

    def _set_dataType(self, value):
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "dataType" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__dataType = value   

    dataType = property(_get_dataType, _set_dataType, None, 
                        "expression value data type")  
    
    @abstractmethod
    def evaluate(self, context):
        """Evaluate the result of the expression in a condition.  Derived 
        classes must implement
        
        @param context: the request context
        @type context: ndg.xacml.core.context.request.Request
        @return: attribute value(s) resulting from execution of this expression
        in a condition
        @rtype: AttributeValue/NoneType
        """
        raise NotImplementedError()


"""NDG XACML ElementTree parsers package 

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "16/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: __init__.py 7109 2010-06-28 12:54:57Z pjkersha $"
import logging
log = logging.getLogger(__name__)

from xml.etree import ElementTree

# Generic ElementTree Helper classes
class QName(ElementTree.QName):
    """Extend ElementTree implementation for improved attribute access support
    """ 

    # ElementTree tag is of the form {namespace}localPart.  getNs extracts the
    # namespace from within the brackets but if not found returns ''
    getNs = staticmethod(lambda tag: getattr(re.search('(?<=\{).+(?=\})', tag),
                                             'group', 
                                             str)())
                                             
    getLocalPart = staticmethod(lambda tag: tag.rsplit('}', 1)[-1])

    def __init__(self, input, tag=None, prefix=None):
        """
        @type input: basestring
        @param input: ElementTree style namespace URI + tag name -
        {namespace URI}tag - OR if tag keyword is set, the namespace URI alone
        @type tag: basestring / None
        @param tag: element tag name.  If None, input must contain the 
        namespace URI and tag name in the ElementTree form {namespace URI}tag.
        @type prefix: basestring / None
        @param prefix: namespace prefix
        """
        
        ElementTree.QName.__init__(self, input, tag=tag)
        
        if tag:
            self.namespaceURI = input
            self.localPart = tag
        else:
            # No tag provided namespace and localPart of QN must be parsed from
            # the namespace
            self.namespaceURI = QName.getNs(input)
            self.localPart = QName.getLocalPart(input)
            
        self.prefix = prefix
    
    def _getPrefix(self):
        """Get prefix
        @return: prefix
        @rtype: string
        """
        return self.__prefix

    def _setPrefix(self, value):
        """Set prefix
        @param value: prefix
        @type value: string
        @raise TypeError: invalid input value type
        """
        self.__prefix = value
    
    prefix = property(_getPrefix, _setPrefix, None, "Prefix")

    def _getLocalPart(self):
        """Get local part
        @return: local part
        @rtype: string
        """
        return self.__localPart
    
    def _setLocalPart(self, value):
        """Set local part
        @param value: local part
        @type value: string
        @raise TypeError: invalid input value type
        """
        self.__localPart = value
        
    localPart = property(_getLocalPart, _setLocalPart, None, "LocalPart")

    def _getNamespaceURI(self):
        """Get namespace URI
        @return: namespace URI
        @rtype: string
        """
        return self.__namespaceURI

    def _setNamespaceURI(self, value):
        """Set namespace URI
        @param value: namespace URI
        @type value: string
        @raise TypeError: invalid input value type
        """
        self.__namespaceURI = value
  
    namespaceURI = property(_getNamespaceURI, _setNamespaceURI, None, 
                            "Namespace URI'")

    def __eq__(self, qname):
        """Enable equality check for QName.  Note that prefixes don't need to
        match
        
        @type qname: ndg.xacml.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        @return: True if input and this object match
        @rtype: bool
        """
        if not isinstance(qname, QName):
            raise TypeError('Expecting %r; got %r' % (QName, type(qname)))
                   
        # Nb. prefixes don't need to agree!         
        return (self.namespaceURI, self.localPart) == \
               (qname.namespaceURI, qname.localPart)

    def __ne__(self, qname):
        """Enable equality check for QName.  Note that prefixes don't need to
        match
        
        @type qname: ndg.xacml.utils.etree.QName
        @param qname: Qualified Name to compare with self 
        """
        return not self.__eq__(qname)



"""NDG XACML Attribute type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: attribute.py 7109 2010-06-28 12:54:57Z pjkersha $"
from ndg.xacml.utils import TypedList
from ndg.xacml.core import XacmlCoreBase
from ndg.xacml.core.attributevalue import AttributeValue


class Attribute(XacmlCoreBase):
    """XACML Attribute type
    
    @cvar ELEMENT_LOCAL_NAME: XML local name for this element
    @type ELEMENT_LOCAL_NAME: string
    @cvar ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME: XML local name for attribute value
    child element
    @type ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME: string
    @cvar DATA_TYPE_ATTRIB_NAME: XML attribute name for data type
    @type DATA_TYPE_ATTRIB_NAME: string
    @cvar ATTRIBUTE_ID_ATTRIB_NAME: attribute ID XML attribute name
    @type ATTRIBUTE_ID_ATTRIB_NAME: string
    @cvar ISSUER_ATTRIB_NAME: issuer XML attribute name
    @type ISSUER_ATTRIB_NAME: string
    
    @ivar __attributeValues: list of attribute values 
    @type __attributeValues: ndg.xacml.utils.TypedList
    @ivar __dataType: data type for this attribute
    @type __dataType: basestring / NoneType
    @ivar __attributeId: identifier for attribute
    @type __attributeId: basestring / NoneType
    @ivar __issuer: issuer id of this attribute
    @type __issuer: basestring / NoneType
    """
    ELEMENT_LOCAL_NAME = 'Attribute'
    ATTRIBUTE_VALUE_ELEMENT_LOCAL_NAME = 'AttributeValue'
    DATA_TYPE_ATTRIB_NAME = 'DataType'
    ATTRIBUTE_ID_ATTRIB_NAME = 'AttributeId'
    ISSUER_ATTRIB_NAME = 'Issuer'
    
    __slots__ = ('__attributeValues', '__dataType', '__attributeId', '__issuer')
    
    def __init__(self):
        self.__attributeValues = TypedList(AttributeValue)
        self.__dataType = None
        self.__attributeId = None
        self.__issuer = None

    @property
    def attributeValues(self):
        """Get attribute values
        
        @return: list of attribute values
        @rtype: ndg.xacml.utils.TypedList
        """
        return self.__attributeValues
     
    @attributeValues.setter
    def attributeValues(self, value):
        """Set attribute values
        
        @param value: list of attribute values 
        @type value: ndg.xacml.utils.TypedList
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, TypedList):
            raise TypeError('Expecting %r type for "attributeValues" '
                            'attribute; got %r' % (TypedList, type(value)))
            
        self.__attributeValues = value
                   
    def _get_dataType(self):
        """Get data type
        @return: attribute data type
        @rtype: basestring / NoneType
        """
        return self.__dataType

    def _set_dataType(self, value):
        """Set data type
        @param value: attribute data type
        @type value: basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "dataType" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__dataType = value   

    dataType = property(_get_dataType, _set_dataType, None, 
                        "Attribute data type") 
        
    @property
    def attributeId(self):
        """Get Attribute Id
        @return: attribute Id
        @rtype: basestring / NoneType
        """
        return self.__attributeId

    @attributeId.setter
    def attributeId(self, value):
        """Set Attribute Id
        @param value: attribute id
        @type value: basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "attributeId" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__attributeId = value  
        
    @property
    def issuer(self):
        """Get Issuer
        @return: attribute issuer
        @rtype: basestring / NoneType
        """
        return self.__issuer

    @issuer.setter
    def issuer(self, value):
        """Set Issuer
        @param value: attribute issuer
        @type value: basestring
        @raise TypeError: incorrect input type
        """
        if not isinstance(value, basestring):
            raise TypeError('Expecting %r type for "issuer" '
                            'attribute; got %r' % (basestring, type(value)))
            
        self.__issuer = value


"""NDG XACML Context Resource type

NERC DataGrid
"""
__author__ = "P J Kershaw"
__date__ = "24/03/10"
__copyright__ = "(C) 2010 Science and Technology Facilities Council"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__license__ = "BSD - see LICENSE file in top-level directory"
__contact__ = "Philip.Kershaw@stfc.ac.uk"
__revision__ = "$Id: resource.py 7098 2010-06-25 13:51:24Z pjkersha $"
from ndg.xacml.core.context import RequestChildBase


class Resource(RequestChildBase):
    """XACML Context Resource type
    
    @cvar ELEMENT_LOCAL_NAME: XML local element name for the resource
    @type ELEMENT_LOCAL_NAME: string
    @cvar RESOURCE_CONTENT_ELEMENT_LOCAL_NAME: XML local element name for the
    resource content
    @type RESOURCE_CONTENT_ELEMENT_LOCAL_NAME: string
    
    @ivar __resourceContent: resource content
    @type __resourceContent: any
    """
    ELEMENT_LOCAL_NAME = 'Resource'
    RESOURCE_CONTENT_ELEMENT_LOCAL_NAME = 'ResourceContent'
    
    __slots__ = ('__resourceContent',)
    
    def __init__(self):
        """Initial resource content instance variable"""
        super(Resource, self).__init__()
        self.__resourceContent = None

    def _get_resourceContent(self):
        """Get resource content
        @return: content
        @rtype: any
        """
        return self.__resourceContent

    def _set_resourceContent(self, value):
        """Set resource content
        @param value: content
        @type value: any
        """
        self.__resourceContent = value   

    resourceContent = property(_get_resourceContent, _set_resourceContent, None, 
                               "Resource content")